import contextlib
from functools import reduce
from threading import BoundedSemaphore
from typing import Dict, List, Optional, Tuple, Iterator, Sequence
from concurrent.futures import Executor

from dlt.common import logger, pendulum
from dlt.common.exceptions import TerminalException
from dlt.common.metrics import LoadJobMetrics
from dlt.common.runtime.signals import sleep
from dlt.common.configuration import with_config, known_sections
from dlt.common.configuration.accessors import config
from dlt.common.pipeline import LoadInfo, LoadMetrics, SupportsPipeline, WithStepInfo
from dlt.common.schema.utils import get_root_table
from dlt.common.storages.load_storage import (
    LoadPackageInfo,
    ParsedLoadJobFileName,
)
from dlt.common.storages.load_package import (
    LoadPackageStateInjectableContext,
    load_package_state as current_load_package,
)
from dlt.common.runners import TRunMetrics, Runnable, workermethod, NullExecutor
from dlt.common.runtime.collector import Collector, NULL_COLLECTOR
from dlt.common.logger import pretty_format_exception
from dlt.common.configuration.container import Container
from dlt.common.schema import Schema
from dlt.common.storages import LoadStorage
from dlt.common.destination import DestinationReference, AnyDestination, Destination
from dlt.common.destination.client import (
    DestinationClientDwhConfiguration,
    HasFollowupJobs,
    JobClientBase,
    WithStagingDataset,
    RunnableLoadJob,
    LoadJob,
    FollowupJobRequest,
    TLoadJobState,
    DestinationClientConfiguration,
    SupportsStagingDestination,
)
from dlt.common.destination.exceptions import (
    DestinationTerminalException,
)
from dlt.common.runtime import signals

from dlt.destinations.job_impl import FinalizedLoadJobWithFollowupJobs

from dlt.load.configuration import LoaderConfiguration
from dlt.load.exceptions import (
    LoadClientJobFailed,
    LoadClientJobRetry,
    LoadClientUnsupportedWriteDisposition,
    LoadClientUnsupportedFileFormats,
    LoadClientJobException,
    FollowupJobCreationFailedException,
    TableChainFollowupJobCreationFailedException,
)
from dlt.load.utils import (
    _extend_tables_with_table_chain,
    get_completed_table_chain,
    init_client,
    filter_new_jobs,
    get_available_worker_slots,
)


class Load(Runnable[Executor], WithStepInfo[LoadMetrics, LoadInfo]):
    pool: Executor

    @with_config(spec=LoaderConfiguration, sections=(known_sections.LOAD,))
    def __init__(
        self,
        destination: AnyDestination,
        staging_destination: AnyDestination = None,
        collector: Collector = NULL_COLLECTOR,
        is_storage_owner: bool = False,
        config: LoaderConfiguration = config.value,
        initial_client_config: DestinationClientConfiguration = config.value,
        initial_staging_client_config: DestinationClientConfiguration = config.value,
    ) -> None:
        self.config = config
        self.collector = collector
        self.initial_client_config = initial_client_config
        self.initial_staging_client_config = initial_staging_client_config
        self.destination = destination
        self.staging_destination = staging_destination
        self.pool = NullExecutor()
        self.load_storage: LoadStorage = self.create_storage(is_storage_owner)
        self._loaded_packages: List[LoadPackageInfo] = []
        self._job_metrics: Dict[str, LoadJobMetrics] = {}
        # job completed signalling event (start blocking)
        self._done_event = BoundedSemaphore()
        self._done_event.acquire()
        self._run_loop_sleep_duration: float = (
            1.0  # amount of time to sleep between querying completed jobs
        )
        super().__init__()

    def create_storage(self, is_storage_owner: bool) -> LoadStorage:
        supported_file_formats = self.destination.capabilities().supported_loader_file_formats
        if self.staging_destination:
            supported_file_formats = (
                self.staging_destination.capabilities().supported_loader_file_formats
            )
        load_storage = LoadStorage(
            is_storage_owner,
            supported_file_formats,
            config=self.config._load_storage_config,
        )
        # add internal job formats
        if issubclass(self.destination.client_class, WithStagingDataset):
            load_storage.supported_job_file_formats += ["sql"]
        if self.staging_destination:
            load_storage.supported_job_file_formats += ["reference"]

        return load_storage

    def get_destination_client(self, schema: Schema) -> JobClientBase:
        return self.destination.client(schema, self.initial_client_config)

    def get_staging_destination_client(self, schema: Schema) -> JobClientBase:
        return self.staging_destination.client(schema, self.initial_staging_client_config)

    def is_staging_destination_job(self, file_path: str) -> bool:
        job_info = ParsedLoadJobFileName.parse(file_path)
        file_type = job_info.file_format
        # for now we know that reference and model jobs always go do the main destination
        if file_type in ["reference", "model"]:
            return False
        return (
            self.staging_destination is not None
            and file_type in self.staging_destination.capabilities().supported_loader_file_formats
        )

    @contextlib.contextmanager
    def maybe_with_staging_dataset(
        self, job_client: JobClientBase, use_staging: bool
    ) -> Iterator[None]:
        """Executes job client methods in context of staging dataset if `table` has `write_disposition` that requires it"""
        if isinstance(job_client, WithStagingDataset) and use_staging:
            with job_client.with_staging_dataset():
                yield
        else:
            yield

    def submit_job(
        self, file_path: str, load_id: str, schema: Schema, restore: bool = False
    ) -> LoadJob:
        job: LoadJob = None

        is_staging_destination_job = self.is_staging_destination_job(file_path)
        job_client = self.get_destination_client(schema)

        # if we have a staging destination and the file is not a reference, send to staging
        active_job_client = (
            self.get_staging_destination_client(schema)
            if is_staging_destination_job
            else job_client
        )

        started_at = pendulum.now()
        try:
            # check file format
            job_info = ParsedLoadJobFileName.parse(file_path)
            if job_info.file_format not in self.load_storage.supported_job_file_formats:
                raise LoadClientUnsupportedFileFormats(
                    job_info.file_format,
                    self.destination.capabilities().supported_loader_file_formats,
                    file_path,
                )
            logger.info(f"Will load file {file_path} with table name {job_info.table_name}")

            # determine which dataset to use
            if is_staging_destination_job:
                use_staging_dataset = isinstance(
                    job_client, SupportsStagingDestination
                ) and job_client.should_load_data_to_staging_dataset_on_staging_destination(
                    job_info.table_name
                )
            else:
                use_staging_dataset = isinstance(
                    job_client, WithStagingDataset
                ) and job_client.should_load_data_to_staging_dataset(job_info.table_name)

            # prepare table to be loaded
            load_table = active_job_client.prepare_load_table(job_info.table_name)
            if load_table["write_disposition"] not in ["append", "replace", "merge"]:
                raise LoadClientUnsupportedWriteDisposition(
                    job_info.table_name, load_table["write_disposition"], file_path
                )
            job = active_job_client.create_load_job(
                load_table,
                self.load_storage.normalized_packages.storage.make_full_path(file_path),
                load_id,
                restore=restore,
            )
            if job is None:
                raise DestinationTerminalException(
                    f"Destination could not create a job for file {file_path}. Typically the file"
                    " extension could not be associated with job type and that indicates an error"
                    " in the code."
                )
        except (TerminalException, AssertionError) as term_ex:
            job = FinalizedLoadJobWithFollowupJobs.from_file_path(
                file_path,
                started_at=started_at,
                status="failed",
                message=pretty_format_exception(),
                exception=term_ex,
            )
        except Exception as retry_ex:
            job = FinalizedLoadJobWithFollowupJobs.from_file_path(
                file_path,
                started_at=started_at,
                status="retry",
                message=pretty_format_exception(),
                exception=retry_ex,
            )

        # move to started jobs in case this is not a restored job
        if not restore:
            job._file_path = self.load_storage.normalized_packages.start_job(
                load_id, job.file_name()
            )

        # only start a thread if this job is runnable
        if isinstance(job, RunnableLoadJob):
            # set job vars
            job.set_run_vars(load_id=load_id, schema=schema, load_table=load_table)
            # submit to pool
            self.pool.submit(Load.w_run_job, *(id(self), job, is_staging_destination_job, use_staging_dataset, schema))  # type: ignore
        else:
            # sanity check: otherwise a job in an actionable state is expected
            assert job.state() in ("completed", "failed", "retry")

        return job

    @staticmethod
    @workermethod
    def w_run_job(
        self: "Load",
        job: RunnableLoadJob,
        use_staging_client: bool,
        use_staging_dataset: bool,
        schema: Schema,
    ) -> None:
        """
        Start a load job in a separate thread
        """
        active_job_client = (
            self.get_staging_destination_client(schema)
            if use_staging_client
            else self.get_destination_client(schema)
        )
        with active_job_client as client:
            with self.maybe_with_staging_dataset(client, use_staging_dataset):
                job.run_managed(active_job_client, self._done_event)

    def start_new_jobs(
        self, load_id: str, schema: Schema, running_jobs: Sequence[LoadJob]
    ) -> Sequence[LoadJob]:
        """
        will retrieve jobs from the new_jobs folder and start as many as there are slots available
        """
        caps = self.destination.capabilities(
            self.destination.configuration(self.initial_client_config)
        )

        # early exit if no slots available
        available_slots = get_available_worker_slots(self.config, caps, running_jobs)
        if available_slots <= 0:
            return []

        # get a list of jobs eligible to be started
        load_files = filter_new_jobs(
            self.load_storage.list_new_jobs(load_id),
            caps,
            self.config,
            running_jobs,
            available_slots,
        )

        logger.info(f"Will load additional {len(load_files)}, creating jobs")
        started_jobs: List[LoadJob] = []
        for file in load_files:
            job = self.submit_job(file, load_id, schema)
            started_jobs.append(job)

        return started_jobs

    def resume_started_jobs(self, load_id: str, schema: Schema) -> List[LoadJob]:
        """
        will check jobs in the started folder and resume them
        """
        jobs: List[LoadJob] = []

        # list all files that were started but not yet completed
        started_jobs = self.load_storage.normalized_packages.list_started_jobs(load_id)

        logger.info(f"{len(started_jobs)} started jobs found, which should be continued")
        if len(started_jobs) == 0:
            return jobs

        for file_path in started_jobs:
            job = self.submit_job(file_path, load_id, schema, restore=True)
            jobs.append(job)

        return jobs

    def get_new_jobs_info(self, load_id: str) -> List[ParsedLoadJobFileName]:
        return [
            ParsedLoadJobFileName.parse(job_file)
            for job_file in self.load_storage.list_new_jobs(load_id)
        ]

    def create_followup_jobs(
        self, load_id: str, state: TLoadJobState, starting_job: LoadJob, schema: Schema
    ) -> None:
        """
        for jobs marked as having followup jobs, find them all and store them to the new jobs folder
        where they will be picked up for execution
        """

        jobs: List[FollowupJobRequest] = []
        if isinstance(starting_job, HasFollowupJobs):
            # check for merge jobs only for jobs executing on the destination, the staging destination jobs must be excluded
            # NOTE: we may move that logic to the interface
            starting_job_file_name = starting_job.file_name()
            if state == "completed" and not self.is_staging_destination_job(starting_job_file_name):
                client = self.destination.client(schema, self.initial_client_config)
                root_job_table = get_root_table(
                    schema.tables, starting_job.job_file_info().table_name
                )
                # if all tables of chain completed, create follow up jobs
                all_jobs_states = self.load_storage.normalized_packages.list_all_jobs_with_states(
                    load_id
                )
                if table_chain := get_completed_table_chain(
                    schema, all_jobs_states, root_job_table, starting_job.job_file_info().job_id()
                ):
                    table_chain_names = [table["name"] for table in table_chain]
                    # all tables will be prepared for main dataset
                    prep_table_chain = [
                        client.prepare_load_table(table_name) for table_name in table_chain_names
                    ]
                    table_chain_jobs = [
                        # we mark all jobs as completed, as by the time the followup job runs the starting job will be in this
                        # folder too
                        self.load_storage.normalized_packages.job_to_job_info(
                            load_id, "completed_jobs", job_state[1]
                        )
                        for job_state in all_jobs_states
                        if job_state[1].table_name in table_chain_names
                        # job being completed is still in started_jobs
                        and job_state[0] in ("completed_jobs", "started_jobs")
                    ]
                    try:
                        if follow_up_jobs := client.create_table_chain_completed_followup_jobs(
                            prep_table_chain, table_chain_jobs
                        ):
                            jobs = jobs + follow_up_jobs
                    except Exception as e:
                        raise TableChainFollowupJobCreationFailedException(
                            root_table_name=prep_table_chain[0]["name"], details=str(e)
                        ) from e

            try:
                jobs = jobs + starting_job.create_followup_jobs(state)
            except Exception as e:
                raise FollowupJobCreationFailedException(job_id=starting_job.job_id()) from e

        # import all followup jobs to the new jobs folder
        for followup_job in jobs:
            # save all created jobs
            self.load_storage.normalized_packages.import_job(
                load_id, followup_job.new_file_path(), job_state="new_jobs"
            )
            logger.info(
                f"Job {starting_job.job_id()} CREATED a new FOLLOWUP JOB"
                f" {followup_job.new_file_path()} placed in new_jobs"
            )
        self.collector.update("Jobs", inc=0, inc_total=len(jobs))

    def complete_jobs(
        self, load_id: str, jobs: Sequence[LoadJob], schema: Schema
    ) -> Tuple[List[LoadJob], List[LoadJob], Optional[LoadClientJobException]]:
        """Run periodically in the main thread to collect job execution statuses.

        After detecting change of status, it commits the job state by moving it to the right folder.
        May create one or more follow-up jobs that get scheduled as new jobs. New jobs are created
        only from jobs in terminal states (completed / failed).

        Args:
            load_id (str): Identifier of the load package to which the jobs belong.
            jobs (Sequence[LoadJob]): Jobs to check and possibly finalize (eg. still running or just finished).
            schema (Schema): Active schema used to resolve clients and create follow-up jobs.

        Returns:
            Tuple[List[LoadJob], List[LoadJob], Optional[LoadClientJobException]]:
                A 3-tuple containing:
                - remaining_jobs: List of jobs still in a non-terminal state ("ready" or "running")
                  that should be checked again in the next loop iteration.
                - finalized_jobs: List of jobs that reached a terminal state during this call
                  ("completed" or "failed") and were moved to the appropriate storage folder.
                - pending_exception: If configured, an exception to be raised by the caller after
                  draining the pool. Set when:
                    * a job failed and `raise_on_failed_jobs` is True -> LoadClientJobFailed
                    * a job hit the retry threshold and `raise_on_max_retries` is set -> LoadClientJobRetry
                  Otherwise None.

        Side Effects:
            - Moves job files between normalized package folders (eg. start -> completed/retry/failed).
            - Caches per-job LoadJobMetrics in `self._job_metrics` for this process.
            - Updates Collector counters (and "Failed" label when a job fails).
            - May import follow-up jobs into the "new_jobs" folder via `create_followup_jobs`.
        """
        # list of jobs still running
        remaining_jobs: List[LoadJob] = []
        # list of jobs in final state
        finalized_jobs: List[LoadJob] = []
        # if an exception condition was met, return it to the main runner
        pending_exception: Optional[LoadClientJobException] = None

        logger.info(f"Will complete {len(jobs)} for {load_id}")
        for ii in range(len(jobs)):
            job = jobs[ii]
            logger.debug(f"Checking state for job {job.job_id()}")
            state: TLoadJobState = job.state()
            if state in ("ready", "running"):
                # ask again
                logger.debug(f"job {job.job_id()} still running")
                remaining_jobs.append(job)
            elif state == "failed":
                # create followup jobs
                self.create_followup_jobs(load_id, state, job, schema)
                # try to get exception message from job
                failed_message = job.failed_message()
                self.load_storage.normalized_packages.fail_job(
                    load_id, job.file_name(), failed_message
                )
                logger.error(
                    f"Job for {job.job_id()} failed terminally in load {load_id} with message"
                    f" {failed_message}"
                )
                # schedule exception on job failure
                if self.config.raise_on_failed_jobs:
                    pending_exception = LoadClientJobFailed(
                        load_id,
                        job.job_file_info().job_id(),
                        failed_message,
                        job.exception(),
                    )
                finalized_jobs.append(job)
            elif state == "retry":
                # try to get exception message from job
                retry_message = job.failed_message()
                # move back to new folder to try again
                self.load_storage.normalized_packages.retry_job(load_id, job.file_name())
                logger.warning(
                    f"Job for {job.job_id()} retried in load {load_id} with message {retry_message}"
                )
                # possibly schedule exception on too many retries
                if self.config.raise_on_max_retries:
                    r_c = job.job_file_info().retry_count + 1
                    if r_c > 0 and r_c % self.config.raise_on_max_retries == 0:
                        pending_exception = LoadClientJobRetry(
                            load_id,
                            job.job_id(),
                            r_c,
                            self.config.raise_on_max_retries,
                            failed_message=retry_message,
                            exception=job.exception(),
                        )
            elif state == "completed":
                # create followup jobs
                self.create_followup_jobs(load_id, state, job, schema)
                # move to completed folder after followup jobs are created
                # in case of exception when creating followup job, the loader will retry operation and try to complete again
                self.load_storage.normalized_packages.complete_job(load_id, job.file_name())
                logger.info(f"Job for {job.job_id()} completed in load {load_id}")
                finalized_jobs.append(job)
            else:
                raise Exception("Incorrect job state")

            # TODO: metrics should be persisted. this is different vs. all other steps because load step
            # may be restarted in the middle of execution
            # NOTE: we could use package state but cases with 100k jobs must be tested
            metrics = job.metrics()
            if metrics:
                self._job_metrics[job.job_id()] = metrics

            if state in ["failed", "completed"]:
                self.collector.update("Jobs")
                if state == "failed":
                    self.collector.update(
                        "Jobs", 1, message="WARNING: Some of the jobs failed!", label="Failed"
                    )

        return remaining_jobs, finalized_jobs, pending_exception

    def complete_package(self, load_id: str, schema: Schema, aborted: bool = False) -> None:
        # do not commit load id for aborted packages
        if not aborted:
            with self.get_destination_client(schema) as job_client:
                with Container().injectable_context(
                    LoadPackageStateInjectableContext(
                        storage=self.load_storage.normalized_packages,
                        load_id=load_id,
                    )
                ):
                    job_client.complete_load(load_id)
                    self._maybe_truncate_staging_dataset(schema, job_client)

        self.load_storage.complete_load_package(load_id, aborted)
        self.gather_metrics(load_id, finished=True)
        # delete jobs only after metrics collected
        self.load_storage.maybe_remove_completed_jobs(load_id)
        logger.info(
            f"All jobs completed, archiving package {load_id} with aborted set to {aborted}"
        )

    def gather_metrics(self, load_id: str, finished: bool) -> None:
        # collect package info
        self._loaded_packages.append(self.load_storage.get_load_package_info(load_id))
        self._step_info_complete_load_id(load_id, finished=finished)

    def initialize_package(
        self, load_id: str, schema: Schema, new_jobs: List[ParsedLoadJobFileName]
    ) -> List[LoadJob]:
        # get dropped and truncated tables that were added in the extract step if refresh was requested
        # NOTE: if naming convention was updated those names correspond to the old naming convention
        # and they must be like that in order to drop existing tables
        dropped_tables = current_load_package()["state"].get("dropped_tables", [])
        truncated_tables = current_load_package()["state"].get("truncated_tables", [])

        # initialize analytical storage ie. create dataset required by passed schema
        with self.get_destination_client(schema) as job_client:
            if (expected_update := self.load_storage.begin_schema_update(load_id)) is not None:
                # init job client
                applied_update = init_client(
                    job_client,
                    schema,
                    new_jobs,
                    expected_update,
                    job_client.should_truncate_table_before_load,
                    (
                        job_client.should_load_data_to_staging_dataset
                        if isinstance(job_client, WithStagingDataset)
                        else None
                    ),
                    lambda table_name: True,  # drop all passed tables
                    drop_tables=dropped_tables,
                    truncate_tables=truncated_tables,
                )

                # init staging client
                if self.staging_destination:
                    assert isinstance(job_client, SupportsStagingDestination), (
                        f"Job client for destination {self.destination.destination_type} does not"
                        " implement SupportsStagingDestination"
                    )

                    with self.get_staging_destination_client(schema) as staging_client:
                        init_client(
                            staging_client,
                            schema,
                            new_jobs,
                            expected_update,
                            # should_truncate_staging,
                            job_client.should_truncate_table_before_load_on_staging_destination,
                            job_client.should_load_data_to_staging_dataset_on_staging_destination,
                            # should we drop tables also on staging destination
                            job_client.should_drop_table_on_staging_destination,
                            drop_tables=dropped_tables,
                            truncate_tables=truncated_tables,
                        )
                self.load_storage.commit_schema_update(load_id, applied_update)

            # collect all unfinished jobs
            return self.resume_started_jobs(load_id, schema)

    def init_jobs_counter(self, load_id: str) -> None:
        # update counter we only care about the jobs that are scheduled to be loaded
        package_jobs = self.load_storage.normalized_packages.get_load_package_jobs(load_id)
        total_jobs = reduce(lambda p, c: p + len(c), package_jobs.values(), 0)
        no_failed_jobs = len(package_jobs["failed_jobs"])
        no_completed_jobs = len(package_jobs["completed_jobs"]) + no_failed_jobs
        self.collector.update("Jobs", no_completed_jobs, total_jobs)
        if no_failed_jobs > 0:
            self.collector.update(
                "Jobs", no_failed_jobs, message="WARNING: Some of the jobs failed!", label="Failed"
            )

    def load_single_package(self, load_id: str, schema: Schema) -> None:
        new_jobs = self.get_new_jobs_info(load_id)
        self.init_jobs_counter(load_id)
        running_jobs = self.initialize_package(load_id, schema, new_jobs)
        # loop until all jobs are processed
        pending_exception: Optional[LoadClientJobException] = None
        while True:
            # we continuously spool new jobs and complete finished ones
            running_jobs, finalized_jobs, new_pending_exception = self.complete_jobs(
                load_id, running_jobs, schema
            )
            pending_exception = pending_exception or new_pending_exception

            # do not spool new jobs if there was a signal or an exception was encountered
            # we inform the users how many jobs remain when shutting down, but only if the count of running jobs
            # has changed (as determined by finalized jobs)
            if signals.was_signal_received() and not self.config.start_new_jobs_on_signal:
                if finalized_jobs:
                    logger.info(
                        f"Signal received, draining running jobs. {len(running_jobs)} to go."
                    )
            elif pending_exception:
                if finalized_jobs:
                    logger.info(
                        f"Exception for job {pending_exception.job_id} received, draining"
                        f" running jobs. {len(running_jobs)} to go."
                    )
            else:
                running_jobs += self.start_new_jobs(load_id, schema, running_jobs)

            # update metrics on each run so they are available even in exception
            metrics: LoadMetrics = {
                "started_at": None,
                "finished_at": None,
                "job_metrics": self._job_metrics,
            }
            self._step_info_update_metrics(load_id, metrics)

            if len(running_jobs) > 0:
                # wakes when managed job enters terminal state (in the pool)
                acquired = self._done_event.acquire(timeout=self._run_loop_sleep_duration)
                logger.debug(f"Sleep on package: {load_id} acquired: {acquired}")
            else:
                break

        remaining_jobs = self.load_storage.list_new_jobs(load_id)
        # if a pending exception was discovered during completion of jobs
        # we can raise it now
        if pending_exception:
            if isinstance(pending_exception, LoadClientJobFailed):
                # the package is completed and skipped
                self.complete_package(load_id, schema, aborted=True)
            else:
                self.gather_metrics(load_id, finished=False)
                logger.warning(
                    f"Package {load_id} was not fully loaded. Load job pool is successfully drained"
                    f" but {len(remaining_jobs)} new jobs are left in the package."
                )
            # raise exception with continuous backtrace into client exception
            raise pending_exception from pending_exception.client_exception

        # pool is drained
        if not remaining_jobs:
            # no new jobs, load package done
            self.complete_package(load_id, schema, aborted=False)
        else:
            self.gather_metrics(load_id, finished=False)
            logger.warning(
                f"Package {load_id} was not fully loaded. Load job pool is successfully drained but"
                f" {len(remaining_jobs)} new jobs are left in the package."
            )

    def run(self, pool: Optional[Executor]) -> TRunMetrics:
        # store pool
        self.pool = pool or NullExecutor()

        logger.info("Running file loading")
        # get list of loads and order by name ASC to execute schema updates
        loads = self.load_storage.list_normalized_packages()
        logger.info(f"Found {len(loads)} load packages")
        if len(loads) == 0:
            return TRunMetrics(True, 0)

        # load the schema from the package
        load_id = loads[0]
        logger.info(f"Loading schema from load package in {load_id}")
        schema = self.load_storage.normalized_packages.load_schema(load_id)
        logger.info(f"Loaded schema name {schema.name} and version {schema.stored_version}")

        # get top load id and mark as being processed
        with self.collector(f"Load {schema.name} in {load_id}"):
            with Container().injectable_context(
                LoadPackageStateInjectableContext(
                    storage=self.load_storage.normalized_packages,
                    load_id=load_id,
                )
            ):
                # the same load id may be processed across multiple runs
                if self.current_load_id is None:
                    self._job_metrics = {}
                    self._step_info_start_load_id(load_id)
                self.load_single_package(load_id, schema)

        return TRunMetrics(False, len(self.load_storage.list_normalized_packages()))

    def _maybe_truncate_staging_dataset(self, schema: Schema, job_client: JobClientBase) -> None:
        """
        Truncate the staging dataset if one used,
        and configuration requests truncation.

        Args:
            schema (Schema): Schema to use for the staging dataset.
            job_client (JobClientBase):
                Job client to use for the staging dataset.
        """
        if not (
            isinstance(job_client, WithStagingDataset) and self.config.truncate_staging_dataset
        ):
            return

        data_tables = schema.data_table_names()
        tables = _extend_tables_with_table_chain(
            schema, data_tables, data_tables, job_client.should_load_data_to_staging_dataset
        )

        try:
            with self.get_destination_client(schema) as client:
                with client.with_staging_dataset():  # type: ignore
                    client.initialize_storage(truncate_tables=tables)

        except Exception as exc:
            logger.warn(
                f"Staging dataset truncate failed due to the following error: {exc}"
                " However, it didn't affect the data integrity."
            )

    def get_step_info(
        self,
        pipeline: SupportsPipeline,
    ) -> LoadInfo:
        # TODO: LoadInfo should hold many datasets
        load_ids = list(self._load_id_metrics.keys())
        metrics: Dict[str, List[LoadMetrics]] = {}
        # get load packages and dataset_name from the last package
        _dataset_name: str = None
        for load_id in load_ids:
            # find package in load packages
            package_info = next((p for p in self._loaded_packages if p.load_id == load_id), None)
            # in rare cases (ie CTRL-C unhandled) package may be not present upfront
            if package_info is None:
                package_info = self.load_storage.get_load_package_info(load_id)
                self._loaded_packages.append(package_info)
            # TODO: each load id may have a separate dataset so construct a list of datasets here
            if isinstance(self.initial_client_config, DestinationClientDwhConfiguration):
                _dataset_name = self.initial_client_config.normalize_dataset_name(
                    package_info.schema
                )
            metrics[load_id] = self._step_info_metrics(load_id)

        return LoadInfo(
            pipeline,
            metrics,
            DestinationReference.normalize_type(self.initial_client_config.destination_type),
            str(self.initial_client_config),
            self.initial_client_config.destination_name
            or Destination.to_name(self.initial_client_config.destination_type),
            self.initial_client_config.environment,
            (
                DestinationReference.normalize_type(
                    self.initial_staging_client_config.destination_type
                )
                if self.initial_staging_client_config
                else None
            ),
            (
                self.initial_staging_client_config.destination_name
                or Destination.to_name(self.initial_staging_client_config.destination_type)
                if self.initial_staging_client_config
                else None
            ),
            str(self.initial_staging_client_config) if self.initial_staging_client_config else None,
            self.initial_client_config.fingerprint(),
            _dataset_name,
            list(load_ids),
            self._loaded_packages,
            pipeline.first_run,
        )



from contextlib import contextmanager
from typing import ClassVar, Iterator

from dlt.common.typing import DictStrAny

from .provider import get_key_name
from .doc import BaseDocProvider


class DictionaryProvider(BaseDocProvider):
    NAME: ClassVar[str] = "Dictionary Provider"

    def __init__(self) -> None:
        super().__init__({})

    @staticmethod
    def get_key_name(key: str, *sections: str) -> str:
        return get_key_name(key, "__", *sections)

    @property
    def name(self) -> str:
        return self.NAME

    @property
    def supports_secrets(self) -> bool:
        return True

    @property
    def supports_sections(self) -> bool:
        return True

    @contextmanager
    def values(self, v: DictStrAny) -> Iterator[None]:
        p_values = self._config_doc
        self._config_doc = v
        yield
        self._config_doc = p_values



import os
from types import ModuleType
from typing import Any, ClassVar, Dict, List, Optional, Union
from abc import ABC, abstractmethod

from dlt.common.configuration.providers.provider import ConfigProvider
from dlt.common.configuration.specs.base_configuration import (
    BaseConfiguration,
    ContainerInjectableContext,
)
from dlt.common.configuration.specs.runtime_configuration import RuntimeConfiguration
from dlt.common.configuration.specs.config_providers_context import ConfigProvidersContainer
from dlt.common.typing import Self
from dlt.common.utils import uniq_id


class RunContextBase(ABC):
    """Describes where `dlt` looks for settings, pipeline working folder. Implementations must be picklable."""

    @abstractmethod
    def __init__(self, run_dir: Optional[str], *args: Any, **kwargs: Any):
        """An explicit run_dir, if None, run_dir should be auto-detected by particular implementation"""

    @property
    @abstractmethod
    def name(self) -> str:
        """Name of the run context. Entities like sources and destinations added to registries when this context
        is active, will be scoped to it. Typically corresponds to Python package name ie. `dlt`.
        """

    @property
    @abstractmethod
    def uri(self) -> str:
        """Uniquely identifies the context. By default it is a combination of `run_dir` and `runtime_kwargs`
        to create file:// uri
        """

    @property
    @abstractmethod
    def global_dir(self) -> str:
        """Directory in which global settings are stored ie ~/.dlt/"""

    @property
    @abstractmethod
    def run_dir(self) -> str:
        """Defines the context working directory, defaults to cwd()"""

    @property
    @abstractmethod
    def local_dir(self) -> str:
        """Defines data dir where local relative dirs and files are created, defaults to run_dir"""

    @property
    @abstractmethod
    def settings_dir(self) -> str:
        """Defines where the current settings (secrets and configs) are located"""

    @property
    @abstractmethod
    def data_dir(self) -> str:
        """Defines where the pipelines working folders are stored."""

    @property
    @abstractmethod
    def module(self) -> Optional[ModuleType]:
        """if run_dir is a top level importable python module, returns it, otherwise return None"""

    @property
    @abstractmethod
    def runtime_kwargs(self) -> Dict[str, Any]:
        """Additional kwargs used to initialize this instance of run context, used for reloading"""

    @abstractmethod
    def initial_providers(self) -> List[ConfigProvider]:
        """Returns initial providers for this context"""

    @abstractmethod
    def initialize_runtime(self, runtime_config: RuntimeConfiguration = None) -> None:
        """Initializes runtime (ie. log, telemetry) using RuntimeConfiguration"""
        pass

    @property
    @abstractmethod
    def runtime_config(self) -> RuntimeConfiguration:
        """Runtime configuration used for initialize_runtime"""
        pass

    @property
    @abstractmethod
    def config(self) -> BaseConfiguration:
        """Returns (optionally resolves) run context configuration"""
        pass

    @abstractmethod
    def get_data_entity(self, entity: str) -> str:
        """Gets path in data_dir where `entity` (ie. `pipelines`, `repos`) are stored"""

    @abstractmethod
    def get_run_entity(self, entity: str) -> str:
        """Gets path in run_dir where `entity` (ie. `sources`, `destinations` etc.) are stored"""

    @abstractmethod
    def get_setting(self, setting_path: str) -> str:
        """Gets path in settings_dir where setting (ie. `secrets.toml`) are stored"""

    @abstractmethod
    def unplug(self) -> None:
        """Called when context removed from container"""

    @abstractmethod
    def plug(self) -> None:
        """Called when context is added to container"""

    def reload(self) -> "RunContextBase":
        """This will reload current context by triggering run context plugin via Container"""
        from dlt.common.configuration.container import Container

        plug_ctx = Container()[PluggableRunContext]
        plug_ctx.reload(self.run_dir, runtime_kwargs=self.runtime_kwargs)
        return plug_ctx.context

    @staticmethod
    def import_run_dir_module(run_dir: str) -> ModuleType:
        """Returns a top Python module of the workspace (if importable)"""
        import importlib

        # trailing separator will be removed by abspath
        run_dir = os.path.abspath(run_dir)
        base_dir = os.path.basename(run_dir)
        if not base_dir:
            raise ImportError(f"`{run_dir=:}` looks like filesystem root")
        m_ = importlib.import_module(base_dir)
        if m_.__file__ and m_.__file__.startswith(run_dir):
            return m_
        else:
            raise ImportError(
                f"`{run_dir=:}` doesn't belong to module `{m_.__file__}` which seems unrelated."
            )

    @abstractmethod
    def reset_config(self) -> None:
        """Hook for contexts that store resolved configuration to reset it"""


class ProfilesRunContext(RunContextBase):
    """Adds profile support on run context. Note: runtime checkable protocols are slow on isinstance"""

    @property
    @abstractmethod
    def profile(self) -> str:
        """Returns current profile name"""

    @property
    @abstractmethod
    def default_profile(self) -> str:
        """Returns default profile name"""

    @abstractmethod
    def available_profiles(self) -> List[str]:
        """Returns available profiles"""

    def configured_profiles(self) -> List[str]:
        """Returns profiles with configurations or dlt entities, same as available by default"""
        return self.available_profiles()

    @abstractmethod
    def switch_profile(self, new_profile: str) -> Self:
        """Switches current profile and returns new run context"""


class PluggableRunContext(ContainerInjectableContext):
    """Injectable run context taken via plugin"""

    global_affinity: ClassVar[bool] = True

    context: RunContextBase = None
    providers: ConfigProvidersContainer

    _context_stack: List[Any] = []

    def __init__(self, init_context: RunContextBase = None) -> None:
        super().__init__()

        if init_context:
            self.context = init_context
        else:
            # autodetect run dir
            self._plug(run_dir=None)
        self.providers = ConfigProvidersContainer(self.context.initial_providers())

    def reload(
        self,
        run_dir_or_context: Optional[Union[str, RunContextBase]] = None,
        runtime_kwargs: Dict[str, Any] = None,
    ) -> None:
        """Reloads the context, using existing settings if not overwritten with method args"""

        if run_dir_or_context is None:
            run_dir_or_context = self.context.run_dir
            if runtime_kwargs is None:
                runtime_kwargs = self.context.runtime_kwargs
            elif self.context.runtime_kwargs:
                runtime_kwargs = {**self.context.runtime_kwargs, **runtime_kwargs}

        self.before_remove()
        if isinstance(run_dir_or_context, str):
            self._plug(run_dir_or_context, runtime_kwargs=runtime_kwargs)
        else:
            self.context = run_dir_or_context
        self.providers = ConfigProvidersContainer(self.context.initial_providers())
        self.after_add()
        # adds remaining providers and initializes runtime
        self.add_extras()

    def reload_providers(self) -> None:
        self.providers = ConfigProvidersContainer(self.context.initial_providers())
        # Re-add extras and re-initialize runtime so changes take effect
        self.providers.add_extras()
        # Invalidate any cached configuration on the context so it re-resolves using new providers
        self.context.reset_config()

    def after_add(self) -> None:
        super().after_add()

        # initialize runtime if context comes back into container
        self.initialize_runtime()

    def before_remove(self) -> None:
        super().before_remove()

        if self.context:
            self.context.unplug()

    def add_extras(self) -> None:
        # add extra providers
        self.providers.add_extras()
        # resolve runtime configuration
        self.initialize_runtime()
        # plug context
        self.context.plug()

    def initialize_runtime(self) -> None:
        """Calls initialize_runtime on context only if active in container. We do not want
        to initialize runtime if instance is not active
        """
        if not self.in_container:
            return
        self.context.initialize_runtime()

    def _plug(self, run_dir: Optional[str], runtime_kwargs: Dict[str, Any] = None) -> None:
        from dlt.common.configuration import plugins

        m = plugins.manager()
        self.context = m.hook.plug_run_context(run_dir=run_dir, runtime_kwargs=runtime_kwargs)
        assert self.context, "plug_run_context hook returned None"

    def push_context(self) -> str:
        """Pushes current context on stack and returns assert cookie"""
        cookie = uniq_id()
        self._context_stack.append((cookie, self.context, self.providers))
        return cookie

    def pop_context(self, cookie: str) -> None:
        """Pops context from stack and re-initializes it if in container"""
        _c, context, providers = self._context_stack.pop()
        if cookie != _c:
            raise ValueError(
                f"Run context stack mangled. Got cookie `{_c}` but expected `{cookie}`"
            )
        self.reload(context)

    def drop_context(self, cookie: str) -> None:
        """Pops context form stack but leaves new context for good"""
        state_ = self._context_stack.pop()
        _c = state_[0]
        if cookie != _c:
            raise ValueError(
                f"Run context stack mangled. Got cookie `{_c}` but expected `{cookie}`"
            )



from copy import deepcopy
from typing import Any, Dict, Literal, Optional, Set, cast

import sqlglot

from dlt.common.schema.typing import TTableSchemaColumns
from dlt.common.schema.utils import merge_columns
from dlt.common.typing import NoneType
from dlt.destinations.impl.clickhouse.typing import (
    CODEC_HINT,
    PARTITION_HINT,
    SETTINGS_HINT,
    SORT_HINT,
    TABLE_ENGINE_TYPES,
    TABLE_ENGINE_TYPE_HINT,
    TColumnCodecs,
    TMergeTreeSettings,
    TSQLExprOrColumnSeq,
    TTableEngineType,
)
from dlt.destinations.utils import get_resource_for_adapter
from dlt.common.exceptions import ValueErrorWithKnownValues
from dlt.extract import DltResource
from dlt.extract.items import TTableHintTemplate


def clickhouse_adapter(
    data: Any,
    table_engine_type: Optional[TTableEngineType] = None,
    sort: Optional[TSQLExprOrColumnSeq] = None,
    partition: Optional[TSQLExprOrColumnSeq] = None,
    settings: Optional[TMergeTreeSettings] = None,
    codecs: Optional[TColumnCodecs] = None,
) -> DltResource:
    """Adapts the given data by applying Clickhouse-specific hints.

    Args:
        data (Any): The data to be transformed. It can be raw data or an instance
            of DltResource. If raw data, the function wraps it into a DltResource
            object.
        table_engine_type (TTableEngineType, optional): The table index type used when creating
            the Clickhouse table.
        sort (TSQLExprOrColumnSeq, optional): Sorting key SQL expression or sequence of column
            names. Used to generated `ORDER BY` clause of table creation statement. If passing a SQL
            expression, use normalized column names when referring to columns.
        partition (TSQLExprOrColumnSeq, optional): Partition key SQL expression or sequence of
            column names. Used to generated `PARTITION BY` clause of table creation statement. If
            passing a SQL expression, use normalized column names when referring to columns.
        settings (TMergeTreeSettings, optional): Dictionary of MergeTree settings to apply to the
            table. Will be added to `SETTINGS` clause of table creation statement.
        codecs (TColumnCodecs, optional): Dictionary of codecs to apply to the table's columns.
            Will be added as `CODEC` clauses in column definitions of table creation statement.

    Returns:
        DltResource: A resource with applied Clickhouse-specific hints.

    Raises:
        ValueError: If input for `table_engine_type` is invalid.
        TypeError: If input types for `sort`, `partition`, `settings`, or `codecs` are invalid.

    Examples:
        Set table engine type:

        >>> data = [{"name": "Alice", "description": "Software Developer"}]
        >>> clickhouse_adapter(data, table_engine_type="merge_tree")

        Set sort and partition keys:

        >>> data = [{"date": "2024-01-01", "town": "Springfield", "street": "Evergreen Terrace"}]
        >>> clickhouse_adapter(
        >>>     data,
        >>>     sort=["town", "street"],  # can also be SQL expression
        >>>     partition="toYYYYMM(date)"  # can also be sequence of column names
        >>> )

        Set MergeTree settings:

        >>> clickhouse_adapter(
        >>>     data,
        >>>     settings={"allow_nullable_key": True, "max_suspicious_broken_parts": 500}
        >>> )

        Set column codecs:

        >>> clickhouse_adapter(
        >>>     data,
        >>>     codecs={"town": "LZ4HC", "street": "Delta, ZSTD(2)"}
        >>> )
    """

    def raise_if_not_none_str_seq(val: Any, name: str) -> None:
        accepted_types = (NoneType, str, list, tuple)
        if not isinstance(val, accepted_types):
            raise TypeError(
                f"`{name}` must be a string or a list or tuple of strings, got"
                f" '{type(val).__name__}'"
            )

    def raise_if_not_none_dict(val: Any, name: str) -> None:
        accepted_types = (NoneType, dict)
        if not isinstance(val, accepted_types):
            raise TypeError(f"`{name}` must be a dictionary, got '{type(val).__name__}'")

    resource = get_resource_for_adapter(data)
    current_columns = cast(TTableSchemaColumns, resource.columns)

    columns = deepcopy(current_columns)
    additional_table_hints: Dict[str, TTableHintTemplate[Any]] = {}

    # table engine type
    if table_engine_type is not None:
        if table_engine_type not in TABLE_ENGINE_TYPES:
            raise ValueErrorWithKnownValues(
                "table_engine_type", table_engine_type, TABLE_ENGINE_TYPES
            )

        additional_table_hints[TABLE_ENGINE_TYPE_HINT] = table_engine_type

    # sort
    raise_if_not_none_str_seq(sort, "sort")
    if sort:
        additional_table_hints[SORT_HINT] = sort
        columns = set_column_hints_from_table_hint(columns, sort, "sort")

    # partition
    raise_if_not_none_str_seq(partition, "partition")
    if partition:
        additional_table_hints[PARTITION_HINT] = partition
        columns = set_column_hints_from_table_hint(columns, partition, "partition")

    # settings
    raise_if_not_none_dict(settings, "settings")
    if settings:
        additional_table_hints[SETTINGS_HINT] = settings

    # codecs
    raise_if_not_none_dict(codecs, "codecs")
    if codecs:
        partial_codec_columns: TTableSchemaColumns = {
            name: {"name": name, CODEC_HINT: codec} for name, codec in codecs.items()  # type: ignore[typeddict-unknown-key]
        }
        columns = merge_columns(columns, partial_codec_columns)

    resource.apply_hints(columns=columns, additional_table_hints=additional_table_hints)
    return resource


def extract_column_names(sql: str) -> Set[str]:
    cols = sqlglot.parse_one(sql, dialect="clickhouse").find_all(sqlglot.exp.Column)
    return {col.name for col in cols}


def get_column_names_from_table_hint(hint: TSQLExprOrColumnSeq) -> Set[str]:
    if isinstance(hint, str):
        # hint is SQL expression; extract column names
        return extract_column_names(hint)
    # hint is sequence of column names; return as set
    return set(hint)


def set_column_hints_from_table_hint(
    columns: TTableSchemaColumns,
    hint: TSQLExprOrColumnSeq,
    hint_name: Literal["sort", "partition"],
) -> TTableSchemaColumns:
    """Sets column hints based on provided table hint.

    Modifies `columns` in place and returns it.

    When it's a `sort` table hint, it sets `sort` column hints.
    When it's a `partition` table hint, it sets `partition` column hints.

    Principles: table hint takes precedence over column hints.

    Rules:
    1. sets/overrides column hint to True for each column in table hint, even if user provided False
    2. sets nullability to False for each column in table hint, unless user provided nullable=True
    3. removes column hint if it's set to True but not in table hint
    4. retains any user-provided nullability
    """

    table_hint_columns = get_column_names_from_table_hint(hint)

    for name in table_hint_columns:
        if name in columns:  # existing column
            # rule 1
            columns[name][hint_name] = True

            # rule 2
            if columns[name].get("nullable") is not True:
                columns[name]["nullable"] = False
        else:  # new column
            # rules 1 and 2
            columns[name] = {"name": name, "nullable": False, hint_name: True}  # type: ignore[misc]

    # rule 3
    for col in columns.values():
        if col.get(hint_name) is True and col["name"] not in table_hint_columns:
            col.pop(hint_name)

    return columns



from typing import Optional

from concurrent.futures import ThreadPoolExecutor


class ManagedThreadPool:
    def __init__(self, thread_name_prefix: str, max_workers: int = 1) -> None:
        self.thread_name_prefix = thread_name_prefix
        self._max_workers = max_workers
        self._thread_pool: Optional[ThreadPoolExecutor] = None

    def _create_thread_pool(self) -> None:
        assert not self._thread_pool, "Thread pool already created"
        self._thread_pool = ThreadPoolExecutor(
            self._max_workers, thread_name_prefix=self.thread_name_prefix
        )

    @property
    def thread_pool(self) -> ThreadPoolExecutor:
        if not self._thread_pool:
            self._create_thread_pool()
        return self._thread_pool

    def stop(self, wait: bool = True) -> None:
        if self._thread_pool:
            self._thread_pool.shutdown(wait=wait)
            self._thread_pool = None



"""Reads files in s3, gs or azure buckets using fsspec and provides convenience resources for chunked reading of various file formats"""
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union

import dlt
from dlt.extract import decorators
from dlt.common.storages.fsspec_filesystem import (
    FileItem,
    FileItemDict,
    fsspec_filesystem,
    glob_files,
)
from dlt.sources import DltResource
from dlt.sources.credentials import FileSystemCredentials

from dlt.sources.filesystem.helpers import (
    AbstractFileSystem,
    FilesystemConfigurationResource,
)
from dlt.sources.filesystem.readers import (
    ReadersSource,
    _read_csv,
    _read_csv_duckdb,
    _read_jsonl,
    _read_parquet,
)
from dlt.sources.filesystem.settings import DEFAULT_CHUNK_SIZE


@decorators.source(_impl_cls=ReadersSource, spec=FilesystemConfigurationResource)
def readers(  # noqa DOC
    bucket_url: str = dlt.secrets.value,
    credentials: Union[FileSystemCredentials, AbstractFileSystem] = dlt.secrets.value,
    file_glob: str = "*",
    kwargs: Optional[Dict[str, Any]] = None,
    client_kwargs: Optional[Dict[str, Any]] = None,
    incremental: Optional[dlt.sources.incremental[Any]] = None,
) -> Tuple[DltResource, ...]:
    """This source provides a few resources that are chunked file readers. Readers can be further parametrized before use
       read_csv(chunksize, **pandas_kwargs)
       read_jsonl(chunksize)
       read_parquet(chunksize)

    Args:
        bucket_url (str): The url to the bucket.
        credentials (Union[FileSystemCredentials, AbstractFileSystem]): The credentials to the filesystem of fsspec `AbstractFilesystem` instance.
        file_glob (str, optional): The filter to apply to the files in glob format. by default lists all files in bucket_url non-recursively
        kwargs: (Optional[Dict[str, Any]], optional): Additional arguments passed to fsspec constructor ie. dict(use_ssl=True) for s3fs
        client_kwargs: (Optional[Dict[str, Any]], optional): Additional arguments passed to underlying fsspec native client ie. dict(verify="public.crt) for botocore
        incremental (Optional[dlt.sources.incremental[Any]]): Defines incremental cursor on listed files, with `modification_date`
            being the most common choice that returns only files created from the previous run.

    Returns:
        Tuple[DltResource, ...]: A tuple of resources that are chunked file readers.
    """
    return (
        filesystem(
            bucket_url,
            credentials,
            file_glob=file_glob,
            kwargs=kwargs,
            client_kwargs=client_kwargs,
            incremental=incremental,
        )
        | dlt.transformer(name="read_csv")(_read_csv),
        filesystem(
            bucket_url,
            credentials,
            file_glob=file_glob,
            kwargs=kwargs,
            client_kwargs=client_kwargs,
            incremental=incremental,
        )
        | dlt.transformer(name="read_jsonl")(_read_jsonl),
        filesystem(
            bucket_url,
            credentials,
            file_glob=file_glob,
            kwargs=kwargs,
            client_kwargs=client_kwargs,
            incremental=incremental,
        )
        | dlt.transformer(name="read_parquet")(_read_parquet),
        filesystem(
            bucket_url,
            credentials,
            file_glob=file_glob,
            kwargs=kwargs,
            client_kwargs=client_kwargs,
            incremental=incremental,
        )
        | dlt.transformer(name="read_csv_duckdb")(_read_csv_duckdb),
    )


@decorators.resource(primary_key="file_url", spec=FilesystemConfigurationResource)
def filesystem(  # noqa DOC
    bucket_url: str = dlt.secrets.value,
    credentials: Union[FileSystemCredentials, AbstractFileSystem] = dlt.secrets.value,
    file_glob: str = "*",
    files_per_page: int = DEFAULT_CHUNK_SIZE,
    extract_content: bool = False,
    kwargs: Optional[Dict[str, Any]] = None,
    client_kwargs: Optional[Dict[str, Any]] = None,
    incremental: Optional[dlt.sources.incremental[Any]] = None,
) -> Iterator[List[FileItem]]:
    """This resource lists files in `bucket_url` using `file_glob` pattern. The files are yielded as FileItem which also
    provide methods to open and read file data. It should be combined with transformers that further process (ie. load files)

    Args:
        bucket_url (str): The url to the bucket.
        credentials (Union[FileSystemCredentials, AbstractFileSystem]): The credentials to the filesystem of fsspec `AbstractFilesystem` instance.
        file_glob (str): The filter to apply to the files in glob format. by default lists all files in bucket_url non-recursively
        files_per_page (int, optional): The number of files to process at once, defaults to 100.
        extract_content (bool, optional): If true, the content of the file will be extracted if
            false it will return a fsspec file, defaults to False.
        kwargs (Optional[Dict[str, Any]]): Additional arguments passed to fsspec constructor ie. dict(use_ssl=True) for s3fs
        client_kwargs (Optional[Dict[str, Any]]): Additional arguments passed to underlying fsspec native client ie. dict(verify="public.crt) for botocore
        incremental (Optional[dlt.sources.incremental[Any]]): Defines incremental cursor on listed files, with `modification_date`
            being the most common choice that returns only files created from the previous run.

    Yields:
        List[FileItem]: The list of files.
    """
    if isinstance(credentials, AbstractFileSystem):
        fs_client = credentials
    else:
        fs_client = fsspec_filesystem(
            bucket_url, credentials, kwargs=kwargs, client_kwargs=client_kwargs
        )[0]

    files_chunk: List[FileItem] = []

    iter_ = glob_files(fs_client, bucket_url, file_glob)

    # if incremental is set with row order, use it to order the results
    # NOTE: fsspec glob for buckets reads all files before running iterator
    #  so below we do not have real batching anyway
    if incremental and incremental.row_order:
        reverse = (incremental.row_order == "asc" and incremental.last_value_func is min) or (
            incremental.row_order == "desc" and incremental.last_value_func is max
        )
        iter_ = iter(
            sorted(
                list(glob_files(fs_client, bucket_url, file_glob)),
                key=lambda f_: f_[incremental.cursor_path],  # type: ignore[literal-required]
                reverse=reverse,
            )
        )

    for file_model in iter_:
        file_dict = FileItemDict(file_model, fs_client)
        if extract_content:
            file_dict["file_content"] = file_dict.read_bytes()
        files_chunk.append(file_dict)  # type: ignore

        # wait for the chunk to be full
        if len(files_chunk) >= files_per_page:
            yield files_chunk
            files_chunk = []
    if files_chunk:
        yield files_chunk


read_csv = decorators.transformer()(_read_csv)
read_jsonl = decorators.transformer()(_read_jsonl)
read_parquet = decorators.transformer()(_read_parquet)
read_csv_duckdb = decorators.transformer()(_read_csv_duckdb)



import sys
import contextlib
from typing import Any, Iterable, Iterator, Optional, ContextManager
import click


ALWAYS_CHOOSE_DEFAULT = False
ALWAYS_CHOOSE_VALUE: Any = None


@contextlib.contextmanager
def always_choose(always_choose_default: bool, always_choose_value: Any) -> Iterator[None]:
    """Temporarily answer all confirmations and prompts with preset values.

    Args:
        always_choose_default: When True, confirm/prompt calls return their default.
        always_choose_value: When set, confirm/prompt calls return this value instead.
    """
    global ALWAYS_CHOOSE_DEFAULT, ALWAYS_CHOOSE_VALUE
    _always_choose_default = ALWAYS_CHOOSE_DEFAULT
    _always_choose_value = ALWAYS_CHOOSE_VALUE
    ALWAYS_CHOOSE_DEFAULT = always_choose_default
    ALWAYS_CHOOSE_VALUE = always_choose_value
    try:
        yield
    finally:
        ALWAYS_CHOOSE_DEFAULT = _always_choose_default
        ALWAYS_CHOOSE_VALUE = _always_choose_value


@contextlib.contextmanager
def suppress_echo() -> Iterator[None]:
    """Temporarily suppress all fmt output."""
    global echo, secho, error, warning, note
    original_echo, original_secho = echo, secho
    original_error, original_warning, original_note = error, warning, note

    def noop(*args: Any, **kwargs: Any) -> None:
        pass

    echo = secho = error = warning = note = noop
    try:
        yield
    finally:
        echo, secho = original_echo, original_secho
        error, warning, note = original_error, original_warning, original_note


def maybe_no_stdin() -> ContextManager[None]:
    """Automatically choose default values if stdin not connected"""
    return always_choose(
        True if not sys.stdin.isatty() else ALWAYS_CHOOSE_DEFAULT, ALWAYS_CHOOSE_VALUE
    )


echo = click.echo
secho = click.secho
style = click.style


def bold(msg: str) -> str:
    return click.style(msg, bold=True, reset=True)


def warning_style(msg: str) -> str:
    return click.style(msg, fg="yellow", reset=True)


def error(msg: str) -> None:
    click.secho("ERROR: " + msg, fg="red")


def warning(msg: str) -> None:
    click.secho("WARNING: " + msg, fg="yellow")


def note(msg: str) -> None:
    click.secho("NOTE: " + msg, fg="green")


def confirm(text: str, default: Optional[bool] = None) -> bool:
    if ALWAYS_CHOOSE_VALUE:
        return bool(ALWAYS_CHOOSE_VALUE)
    if ALWAYS_CHOOSE_DEFAULT:
        assert default is not None
        return default
    return click.confirm(text, default=default)


def prompt(
    text: str,
    choices: Iterable[str],
    default: Optional[Any] = None,
    show_choices: bool = True,
    show_default: bool = True,
) -> Any:
    if ALWAYS_CHOOSE_VALUE:
        assert ALWAYS_CHOOSE_VALUE in choices
        return ALWAYS_CHOOSE_VALUE
    if ALWAYS_CHOOSE_DEFAULT:
        assert default is not None
        return default
    click_choices = click.Choice(choices)
    return click.prompt(
        text,
        type=click_choices,
        default=default,
        show_choices=show_choices,
        show_default=show_default,
    )


def text_input(text: str) -> str:
    return click.prompt(text)  # type: ignore[no-any-return]



from typing import Dict

from pydantic import Field
from fastmcp.exceptions import ToolError

from dlt.common.typing import Annotated
from dlt._workspace.cli.ai.utils import (
    DEFAULT_AI_WORKBENCH_BRANCH,
    DEFAULT_AI_WORKBENCH_REPO,
    fetch_workbench_base,
    fetch_workbench_toolkit_info,
    fetch_workbench_toolkits,
)
from dlt._workspace.mcp.context import with_mcp_tool_telemetry
from dlt._workspace.typing import TToolkitInfo, TWorkbenchToolkitInfo


@with_mcp_tool_telemetry()
def list_toolkits() -> Dict[str, TToolkitInfo]:
    """List available dlt AI toolkits with their names and descriptions."""
    try:
        base = fetch_workbench_base(DEFAULT_AI_WORKBENCH_REPO, DEFAULT_AI_WORKBENCH_BRANCH)
    except FileNotFoundError as ex:
        raise ToolError(str(ex))
    toolkits, _ = fetch_workbench_toolkits(base, listed_only=True)
    return toolkits


@with_mcp_tool_telemetry()
def toolkit_info(
    name: Annotated[str, Field(description="Name of the toolkit to inspect")],
) -> TWorkbenchToolkitInfo:
    """Show detailed contents of a dlt AI toolkit."""
    try:
        info = fetch_workbench_toolkit_info(
            name, DEFAULT_AI_WORKBENCH_REPO, DEFAULT_AI_WORKBENCH_BRANCH
        )
    except (FileNotFoundError, ValueError) as ex:
        raise ToolError(str(ex))
    if info is None:
        raise ToolError("Toolkit '%s' not found" % name)
    return info


__tools__ = (list_toolkits, toolkit_info)



import re
from pathlib import PurePosixPath
from typing import List, Dict, Any, Tuple, Union, Callable, Iterable
from urllib.parse import urlparse

from requests import Response

from .paginators import (
    BasePaginator,
    HeaderLinkPaginator,
    JSONLinkPaginator,
    JSONResponseCursorPaginator,
    SinglePagePaginator,
    PageNumberPaginator,
)

RECORD_KEY_PATTERNS = frozenset(
    [
        "data",
        "items",
        "results",
        "entries",
        "records",
        "rows",
        "entities",
        "payload",
        "content",
        "objects",
        "values",
    ]
)

NON_RECORD_KEY_PATTERNS = frozenset(
    [
        "meta",
        "metadata",
        "pagination",
        "links",
        "extras",
        "headers",
    ]
)

NEXT_PAGE_KEY_PATTERNS = frozenset(["next"])
NEXT_PAGE_DICT_KEY_PATTERNS = frozenset(["href", "url"])
TOTAL_PAGES_KEYS = frozenset(["^total_pages$", "^pages$", "^totalpages$"])


def single_entity_path(path: str) -> bool:
    """Checks if path ends with path param indicating that single object is returned"""
    # get last path segment
    name = PurePosixPath(path).name
    # alphabet for a name taken from https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#fixed-fields-6
    return re.search(r"\{([a-zA-Z0-9\.\-_]+)\}", name) is not None


def matches_any_pattern(key: str, patterns: Iterable[str]) -> bool:
    normalized_key = key.lower()
    return any(re.match(pattern, normalized_key) for pattern in patterns)


def find_all_lists(
    dict_: Dict[str, Any],
    path: Tuple[str, ...] = (),
    result: List[Tuple[Tuple[str, ...], List[Any]]] = None,
) -> List[Tuple[Tuple[str, ...], List[Any]]]:
    """Recursively looks for lists in dict_ and returns tuples
    in format (dictionary keys, list)
    """
    if len(path) > 2:
        return None

    for key, value in dict_.items():
        if isinstance(value, list):
            result.append(((*path, key), value))
        elif isinstance(value, dict):
            find_all_lists(value, path=(*path, key), result=result)

    return result


def find_response_page_data(
    response: Union[Dict[str, Any], List[Any], Any],
) -> Tuple[Tuple[str, ...], Any]:
    """Finds a path to response data, assuming that data is a list, returns a tuple(path, data)"""
    # when a list was returned (or in rare case a simple type or null)
    if not isinstance(response, dict):
        return (("$",), response)
    lists = find_all_lists(response, result=[])
    if len(lists) == 0:
        # could not detect anything
        return (("$",), response)
    # we are ordered by nesting level, find the most suitable list
    try:
        return next(
            list_info
            for list_info in lists
            if list_info[0][-1] in RECORD_KEY_PATTERNS
            and list_info[0][-1] not in NON_RECORD_KEY_PATTERNS
        )
    except StopIteration:
        # return the least nested element
        return lists[0]


def find_next_page_path(
    response: Dict[str, Any], path: Tuple[str, ...] = ()
) -> Tuple[Tuple[str, ...], Any]:
    if not isinstance(response, dict):
        return (None, None)

    for key, value in response.items():
        if matches_any_pattern(key, NEXT_PAGE_KEY_PATTERNS):
            if isinstance(value, dict):
                for dict_key, dict_value in value.items():
                    if matches_any_pattern(dict_key, NEXT_PAGE_DICT_KEY_PATTERNS) and isinstance(
                        dict_value, (str, int, float)
                    ):
                        return ((*path, key, dict_key), dict_value)
            else:
                if isinstance(value, (str, int, float)):
                    return ((*path, key), value)

        if isinstance(value, dict):
            result = find_next_page_path(value, (*path, key))
            if result != (None, None):
                return result

    return (None, None)


def find_total_pages_path(
    response: Dict[str, Any], path: Tuple[str, ...] = ()
) -> Tuple[Tuple[str, ...], Any]:
    if not isinstance(response, dict):
        return (None, None)

    for key, value in response.items():
        if matches_any_pattern(key, TOTAL_PAGES_KEYS) and isinstance(value, (str, int, float)):
            assert key != "pageSize"
            return ((*path, key), value)

        if isinstance(value, dict):
            result = find_total_pages_path(value, (*path, key))
            if result != (None, None):
                return result

    return (None, None)


def header_links_detector(response: Response) -> Tuple[HeaderLinkPaginator, float]:
    links_next_key = "next"

    if response.links.get(links_next_key):
        return HeaderLinkPaginator(), 1.0
    return None, None


def json_links_detector(response: Response) -> Tuple[JSONLinkPaginator, float]:
    dictionary = response.json()
    next_path_parts, next_href = find_next_page_path(dictionary)

    if not next_path_parts:
        return None, None

    try:
        urlparse(next_href)
        if next_href.startswith("http") or next_href.startswith("/"):
            return JSONLinkPaginator(next_url_path=".".join(next_path_parts)), 1.0
    except Exception:
        pass

    return None, None


def cursor_paginator_detector(response: Response) -> Tuple[JSONResponseCursorPaginator, float]:
    dictionary = response.json()
    cursor_path_parts, _ = find_next_page_path(dictionary)

    if not cursor_path_parts:
        return None, None

    return JSONResponseCursorPaginator(cursor_path=".".join(cursor_path_parts)), 0.5


def pages_number_paginator_detector(response: Response) -> Tuple[PageNumberPaginator, float]:
    total_pages_path, total_pages = find_total_pages_path(response.json())
    if not total_pages_path:
        return None, None

    try:
        int(total_pages)
        return PageNumberPaginator(total_path=".".join(total_pages_path)), 0.5
    except Exception:
        pass

    return None, None


def single_page_detector(response: Response) -> Tuple[SinglePagePaginator, float]:
    """This is our fallback paginator, also for results that are single entities"""
    return SinglePagePaginator(), 0.0


class PaginatorFactory:
    def __init__(self, detectors: List[Callable[[Response], Tuple[BasePaginator, float]]] = None):
        """`detectors` are functions taking Response as input and returning paginator instance and
        detection score. Score value:
        1.0 - perfect detection
        0.0 - fallback detection
        in between - partial detection, several paginator parameters are defaults
        """
        if detectors is None:
            detectors = [
                header_links_detector,
                json_links_detector,
                pages_number_paginator_detector,
                cursor_paginator_detector,
                single_page_detector,
            ]
        self.detectors = detectors

    def create_paginator(self, response: Response) -> Tuple[BasePaginator, float]:
        for detector in self.detectors:
            paginator, score = detector(response)
            if paginator:
                return paginator, score
        return None, None



"""Pipeline info, config resolution, destination details, and misc pipeline helpers."""

import os
import platform
import shutil
import subprocess
from typing import Any, Dict, List, Optional, Tuple

from dlt._workspace.helpers.dashboard.typing import TNameValueItem, TPipelineListItem

import dlt
import marimo as mo

from dlt.common.configuration import resolve_configuration
from dlt.common.configuration.specs import known_sections
from dlt.common.destination.client import DestinationClientConfiguration
from dlt.common.storages.configuration import WithLocalFiles
from dlt.common.configuration.exceptions import ConfigFieldMissingException
from dlt.common.utils import map_nested_keys_in_place

from dlt._workspace.cli import utils as cli_utils
from dlt._workspace.helpers.dashboard import strings
from dlt._workspace.helpers.dashboard.config import DashboardConfiguration
from dlt._workspace.helpers.dashboard.utils.formatters import (
    dict_to_table_items,
    format_exception_message,
)
from dlt._workspace.helpers.dashboard.utils.schema import schemas_to_table_items
from dlt._workspace.helpers.dashboard.utils import ui
from dlt.pipeline.exceptions import PipelineConfigMissing
from dlt.pipeline.trace import PipelineTrace


def get_dashboard_config_sections(p: Optional[dlt.Pipeline]) -> Tuple[str, ...]:
    """Find dashboard config section layout for a particular pipeline or for active
    run context type.
    """
    sections: Tuple[str, ...] = ()

    if p is None:
        # use workspace section layout
        context = dlt.current.run_context()
        if context.config is None or not context.config.__class__.__recommended_sections__:
            pass
        else:
            sections = tuple(context.config.__class__.__recommended_sections__) + sections
    else:
        # pipeline section layout
        sections = (known_sections.PIPELINES, p.pipeline_name) + sections

    return sections


def resolve_dashboard_config(p: Optional[dlt.Pipeline]) -> DashboardConfiguration:
    """Resolve the dashboard configuration"""
    return resolve_configuration(
        DashboardConfiguration(),
        sections=get_dashboard_config_sections(p),
    )


def get_pipeline(pipeline_name: str, pipelines_dir: str) -> dlt.Pipeline:
    """Get a pipeline by name. Attach exceptions must be handled by the caller."""
    p = dlt.attach(pipeline_name, pipelines_dir=pipelines_dir)
    p.config.use_single_dataset = False
    return p


def get_destination_config(pipeline: dlt.Pipeline) -> DestinationClientConfiguration:
    """Get the destination config of a pipeline."""
    # NOTE: this uses internal interfaces for now...
    return pipeline.dataset().destination_client.config


def pipeline_details(
    c: DashboardConfiguration, pipeline: dlt.Pipeline, pipelines_dir: str
) -> List[TNameValueItem]:
    """Get the details of a pipeline as name/value table items."""
    try:
        credentials = str(get_destination_config(pipeline).credentials)
    except Exception:
        credentials = strings.overview_credentials_error

    last_executed = strings.overview_no_trace
    if (trace := pipeline.last_trace) and hasattr(trace, "started_at"):
        last_executed = cli_utils.date_from_timestamp_with_ago(trace.started_at, c.datetime_format)

    details_dict = {
        "pipeline_name": pipeline.pipeline_name,
        "destination": (
            pipeline.destination.destination_description
            if pipeline.destination
            else strings.overview_no_destination
        ),
        "last executed": last_executed,
        "credentials": credentials,
        "dataset_name": pipeline.dataset_name,
        "working_dir": pipeline.working_dir,
        "state_version": (
            pipeline.state["_state_version"] if pipeline.state else strings.overview_no_state
        ),
    }

    table_items = dict_to_table_items(details_dict)
    table_items += schemas_to_table_items(pipeline.schemas.values(), pipeline.default_schema_name)
    return table_items


def remote_state_details(pipeline: dlt.Pipeline) -> List[TNameValueItem]:
    """Get the remote state details of a pipeline."""
    error_details = ""
    remote_state = None
    try:
        remote_state = pipeline._restore_state_from_destination()
    except Exception as exc:
        error_details = format_exception_message(exc)

    if not remote_state:
        return dict_to_table_items(
            {"Info": strings.overview_remote_state_error, "Details": error_details}
        )
    remote_schemas = pipeline._get_schemas_from_destination(
        remote_state["schema_names"], always_download=True
    )

    table_items = dict_to_table_items({"state_version": remote_state["_state_version"]})
    table_items += schemas_to_table_items(remote_schemas, pipeline.default_schema_name)
    return table_items


def get_local_data_path(pipeline: dlt.Pipeline) -> str:
    """Get the local data path of a pipeline"""
    if not pipeline.destination:
        return None
    try:
        config = pipeline._get_destination_clients(dlt.Schema("temp"))[0].config
        if isinstance(config, WithLocalFiles):
            return config.local_dir
    except (PipelineConfigMissing, ConfigFieldMissingException):
        pass
    return None


def open_local_folder(folder: str) -> None:
    """Open a folder in the file explorer"""
    system = platform.system()
    if system == "Windows":
        os.startfile(folder)  # type: ignore[attr-defined,unused-ignore]
    elif system == "Darwin":
        subprocess.run(["open", folder], check=True)
    elif shutil.which("wslview"):
        subprocess.run(["wslview", folder], check=True)
    else:
        subprocess.run(["xdg-open", folder], check=True)


def pipeline_link_list(config: DashboardConfiguration, pipelines: List[TPipelineListItem]) -> str:
    """Build a markdown list of links to pipelines."""
    if not pipelines:
        return strings.overview_no_pipelines

    count = 0
    link_list: str = ""
    for _p in pipelines:
        link = f"* [{_p['name']}](?pipeline={_p['name']})"
        link = (
            link
            + strings.overview_last_executed_label
            + cli_utils.date_from_timestamp_with_ago(_p["timestamp"], config.datetime_format)
        )

        link_list += f"{link}\n"
        count += 1
        if count == 10:
            break

    return link_list


def exception_section(p: dlt.Pipeline) -> List[mo.Html]:
    """Build an exception section for a pipeline"""
    if not p or not p.last_trace:
        return []

    if not (exception_step := next((s for s in p.last_trace.steps if s.step_exception), None)):
        return []

    last_exception = exception_step.exception_traces[-1]
    title = f"Exception encountered during last run in step **{exception_step.step}**"
    subtitle = f"{last_exception['exception_type']}: **{last_exception['message']}**"

    _result = []
    _result.append(
        ui.title_and_subtitle(
            title,
            subtitle,
        )
    )

    _exception_traces = []
    for trace in reversed(exception_step.exception_traces):
        _exception_traces.extend(trace["stack_trace"])
        _exception_traces.append(f"{trace['exception_type']}: {trace['message']}")
        _exception_traces.append("\n")
        _exception_traces.append("\n")

    _result.append(
        mo.accordion(
            {
                strings.error_show_full_stacktrace: mo.ui.code_editor(
                    "".join(_exception_traces),
                    language="python",
                    disabled=True,
                    show_copy_button=True,
                )
            },
            lazy=True,
        )
    )
    return [mo.callout(mo.vstack(_result), kind="danger")]


def sanitize_trace_for_display(trace: PipelineTrace) -> Dict[str, Any]:
    """Sanitize a trace for display by cleaning up non-primitive keys."""
    if not trace:
        return {}

    def _remove_non_primitives(obj: Any) -> Any:
        if not isinstance(obj, (str, bool, int, float)):
            return repr(obj)
        return obj

    return map_nested_keys_in_place(_remove_non_primitives, trace.asdict())



from typing import Optional, Dict, Any, cast

from dlt.common.utils import without_none
from dlt.common.exceptions import MissingDependencyException
from dlt.common.typing import TSecretStrValue, DictStrAny, Self
from dlt.common.configuration.specs import (
    CredentialsConfiguration,
    CredentialsWithDefault,
    configspec,
)
from dlt.common.configuration.specs.mixins import WithObjectStoreRsCredentials, WithPyicebergConfig
from dlt.common.configuration.specs.exceptions import (
    InvalidBoto3Session,
    ObjectStoreRsCredentialsException,
)
from dlt import version


@configspec
class AwsCredentialsWithoutDefaults(
    CredentialsConfiguration, WithObjectStoreRsCredentials, WithPyicebergConfig
):
    # credentials without boto implementation
    aws_access_key_id: str = None
    aws_secret_access_key: TSecretStrValue = None
    aws_session_token: Optional[TSecretStrValue] = None
    profile_name: Optional[str] = None
    region_name: Optional[str] = None
    endpoint_url: Optional[str] = None
    s3_url_style: Optional[str] = None
    """Only needed for duckdb sql_client s3 access, for minio this needs to be set to path for example."""

    def to_s3fs_credentials(self) -> Dict[str, Optional[str]]:
        """Dict of keyword arguments that can be passed to s3fs"""
        credentials: DictStrAny = dict(
            key=self.aws_access_key_id,
            secret=self.aws_secret_access_key,
            token=self.aws_session_token,
            profile=self.profile_name,
            endpoint_url=self.endpoint_url,
        )
        if self.region_name:
            credentials["client_kwargs"] = {"region_name": self.region_name}
        if self.endpoint_url:
            # NOTE: we need to make checksum validation optional for boto to work with s3 compat mode
            # https://www.beginswithdata.com/2025/05/14/aws-s3-tools-with-gcs/
            credentials["config_kwargs"] = {
                "response_checksum_validation": "when_required",
                "request_checksum_calculation": "when_required",
            }
        return credentials

    def to_native_representation(self) -> Dict[str, Optional[str]]:
        """Return a dict that can be passed as kwargs to boto3 session"""
        return dict(self)

    def to_session_credentials(self) -> Dict[str, str]:
        return dict(
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
            aws_session_token=self.aws_session_token,
        )

    def to_object_store_rs_credentials(self) -> Dict[str, str]:
        # https://docs.rs/object_store/latest/object_store/aws
        # NOTE: delta rs will set the values below in env variables of the current process
        # https://github.com/delta-io/delta-rs/blob/bdf1c4e765ca457e49d4fa53335d42736220f57f/rust/src/storage/s3.rs#L257
        creds = cast(
            Dict[str, str],
            without_none(
                dict(
                    aws_access_key_id=self.aws_access_key_id,
                    aws_secret_access_key=self.aws_secret_access_key,
                    aws_session_token=self.aws_session_token,
                    region=self.region_name,
                    endpoint_url=self.endpoint_url,
                )
            ),
        )

        if not self.endpoint_url:  # AWS S3
            if not self.region_name:
                raise ObjectStoreRsCredentialsException(
                    "`object_store` Rust crate requires AWS region when using AWS S3."
                )
        else:  # S3-compatible, e.g. MinIO
            if self.endpoint_url.startswith("http://"):
                creds["aws_allow_http"] = "true"

        return creds

    def to_pyiceberg_fileio_config(self) -> Dict[str, Any]:
        return {
            "s3.access-key-id": self.aws_access_key_id,
            "s3.secret-access-key": self.aws_secret_access_key,
            "s3.session-token": self.aws_session_token,
            "s3.region": self.region_name,
            "s3.endpoint": self.endpoint_url,
            "s3.connect-timeout": 300,
        }

    @classmethod
    def from_pyiceberg_fileio_config(cls, file_io: Dict[str, Any]) -> Self:
        credentials: Self = cls()
        credentials.aws_access_key_id = file_io.get("s3.access-key-id")
        credentials.aws_secret_access_key = file_io.get("s3.secret-access-key")
        credentials.aws_session_token = file_io.get("s3.session-token")
        credentials.region_name = file_io.get("s3.region")
        credentials.endpoint_url = file_io.get("s3.endpoint")
        return credentials


@configspec
class AwsCredentials(AwsCredentialsWithoutDefaults, CredentialsWithDefault):
    def on_partial(self) -> None:
        # Try get default credentials
        session = self._to_botocore_session()
        if self._from_session(session) and not self.is_partial():
            self.resolve()

    def to_session_credentials(self) -> Dict[str, str]:
        """Return configured or new aws session token"""
        if self.aws_session_token and self.aws_access_key_id and self.aws_secret_access_key:
            return dict(
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
                aws_session_token=self.aws_session_token,
            )
        sess = self._to_botocore_session()
        client = sess.create_client("sts")
        token = client.get_session_token()
        return dict(
            aws_access_key_id=token["Credentials"]["AccessKeyId"],
            aws_secret_access_key=token["Credentials"]["SecretAccessKey"],
            aws_session_token=token["Credentials"]["SessionToken"],
        )

    def _to_botocore_session(self) -> Any:
        """Creates a botocore session out of the credentials. profile and region will be applied
        as config variables.
        """
        try:
            import botocore.session
            from botocore.config import Config
        except ModuleNotFoundError:
            raise MissingDependencyException(
                self.__class__.__name__, [f"{version.DLT_PKG_NAME}[s3]"]
            )

        # taken from boto3 Session
        session = botocore.session.get_session()
        if self.profile_name is not None:
            session.set_config_variable("profile", self.profile_name)

        if self.aws_access_key_id or self.aws_secret_access_key or self.aws_session_token:
            session.set_credentials(
                self.aws_access_key_id, self.aws_secret_access_key, self.aws_session_token
            )
        if self.region_name is not None:
            session.set_config_variable("region", self.region_name)

        if self.endpoint_url:
            cfg = Config(
                request_checksum_calculation="when_required",
                response_checksum_validation="when_required",
            )
            session.set_default_client_config(cfg)
        return session

    def _from_session(self, session: Any) -> Any:
        """Sets the credentials properties from botocore or boto3 `session` and return session's credentials if found"""
        import botocore.session

        if not isinstance(session, botocore.session.Session):
            # assume this is boto3 session
            session = session._session
        # NOTE: we do not set profile name from boto3 session
        # we either pass it explicitly in `_to_session` so we know it is identical
        # this is what boto3 does: return self._session.profile or 'default' which is obviously wrong (returning default when there's no session)
        self.region_name = session.get_config_variable("region")
        default = session.get_credentials()
        if not default:
            return None
        self.aws_access_key_id = default.access_key
        self.aws_secret_access_key = TSecretStrValue(default.secret_key)
        self.aws_session_token = cast(TSecretStrValue, default.token)
        return default

    def to_native_credentials(self) -> Optional[Any]:
        return self._to_botocore_session().get_credentials()

    def parse_native_representation(self, native_value: Any) -> None:
        """Import external boto3 session"""
        try:
            if self._from_session(native_value):
                self.__is_resolved__ = True
        except Exception:
            raise InvalidBoto3Session(self.__class__, native_value)

    @classmethod
    def from_session(cls, botocore_session: Any) -> "AwsCredentials":
        self = cls()
        self.parse_native_representation(botocore_session)
        return self



import os
import textwrap
from typing import Any, Dict, List, Type, Tuple, NamedTuple, Sequence, Union

from dlt.common.exceptions import DltException, TerminalException
from dlt.common.utils import main_module_file_path


class LookupTrace(NamedTuple):
    provider: str
    sections: Sequence[str]
    key: str
    value: Any


class LookupTraces(NamedTuple):
    """A set of traces generated by embedded config, recursive"""

    spec_name: str
    """type name of the configuration"""
    resolved_fields_set: List[str]
    """list of resolved fields, excluding default values"""
    union_pos: int
    union_count: int
    """Position and total count if union of configs"""
    traces: "FieldLookupTraces"
    """Dict of value traces or embedded config traces"""


NestedLookupTraces = List[Union[LookupTrace, LookupTraces]]
FieldLookupTraces = Dict[str, NestedLookupTraces]


class ConfigurationException(DltException, TerminalException):
    pass


class ConfigurationValueError(ConfigurationException, ValueError):
    pass


class ContainerException(DltException):
    """base exception for all exceptions related to injectable container"""


class ConfigProviderException(ConfigurationException):
    def __init__(self, provider_name: str, *args: Any) -> None:
        self.provider_name = provider_name
        super().__init__(*args)


class ConfigurationWrongTypeException(ConfigurationException):
    def __init__(self, _typ: type) -> None:
        super().__init__(
            f"Invalid configuration instance type `{_typ}`. Configuration instances must derive"
            " from BaseConfiguration and must be decorated with @configspec."
        )


class ConfigFieldMissingException(KeyError, ConfigurationException):
    """raises when not all required config fields are present"""

    def __init__(self, config: Any, traces: FieldLookupTraces) -> None:
        self.traces = traces
        self.config = config
        self.spec_name = type(config).__name__
        self.resolved_fields_set = config.__resolved_fields_set__
        self.fields = list(traces.keys())
        super().__init__(self.spec_name)

    @staticmethod
    def _build_config_error_message(
        # spec_name: str, union_pos: int, union_count: int,
        spec_name: str,
        resolved_keys: List[str],
        traces: FieldLookupTraces,
    ) -> str:
        """Takes in the configuration spec, the resolved key and the configuration resolution traces
        to create a detailed exception / log message about failed configuration resolution.
        """
        msg = ""
        for f, field_traces in traces.items():
            for idx, trace in enumerate(field_traces):
                if isinstance(trace, LookupTraces):
                    fields = list(trace.traces.keys())
                    msg += (
                        f"field `{f}` is itself a configuration `{trace.spec_name}` with missing"
                        f" field(s): {', '.join(f'`{f}`' for f in fields)}"
                    )
                    if trace.union_count > 0:
                        msg += f" (Union {trace.union_pos} of {trace.union_count})"
                    msg += "\n"
                    msg += textwrap.indent(
                        ConfigFieldMissingException._build_config_error_message(
                            trace.spec_name, trace.resolved_fields_set, trace.traces
                        ),
                        "  ",
                    )
                    msg += "\n"
                else:
                    if idx == 0:
                        msg += (
                            f"for field `{f}` the following (config provider, key) were tried in"
                            " order:\n"
                        )
                    msg += f"  ({trace.provider}, {trace.key})\n"
        if resolved_keys:
            msg += (
                f"NOTE: following fields in `{spec_name}` got resolved:"
                f" {', '.join(f'`{f}`' for f in resolved_keys)}\n"
            )
        return msg

    def __str__(self) -> str:
        msg = (
            f"Missing {len(self.fields)} field(s) in configuration `{self.spec_name}`:"
            f" {', '.join(f'`{f}`' for f in self.fields)}\n"
        )
        msg += self._build_config_error_message(
            self.spec_name, self.resolved_fields_set, self.traces
        )

        from dlt.common.configuration.container import Container
        from dlt.common.configuration.specs import PluggableRunContext

        # print locations for config providers
        providers = Container()[PluggableRunContext].providers
        for provider in providers.providers:
            if provider.present_locations:
                locations = "\n".join(
                    [f"\t- {os.path.abspath(loc)}" for loc in provider.present_locations]
                )
                msg += f"Provider `{provider.name}` loaded values from locations:\n{locations}\n"

            # inform on locations that were not found
            not_found_locations = set(provider.locations).difference(provider.present_locations)
            if not_found_locations:
                locations = "\n".join(
                    [f"\t- {os.path.abspath(loc)}" for loc in not_found_locations]
                )
                msg += (
                    f"Provider `{provider.name}` probed but not found the following"
                    f" locations:\n{locations}\n"
                )

            if provider.is_empty:
                msg += (
                    f"WARNING: provider `{provider.name}` is empty. Locations (i.e., files) are"
                    " missing or empty.\n"
                )
        # get pipeline context warning
        msg += get_run_context_warning(main_module_file_path())
        msg += "Learn more: https://dlthub.com/docs/general-usage/credentials/\n"
        return msg

    def attrs(self) -> Dict[str, Any]:
        """Get exception attrs with flattened traces and trace values removed"""
        attrs_ = super().attrs()
        if "traces" in attrs_:
            all_traces: Dict[str, List[LookupTrace]] = {}
            for f, traces in self.traces.items():
                all_traces[f] = flatten_traces(traces, remove_values=True)
            attrs_["traces"] = all_traces
        del attrs_["config"]
        return attrs_

    def was_partially_resolved(self, key: str) -> bool:
        """Checks traces if key was partially resolved"""
        traces = self.traces[key]
        resolved_fields_count = 0
        # count resolved fields
        for trace in traces:
            if isinstance(trace, LookupTraces):
                resolved_fields_count += len(trace.resolved_fields_set or [])
        return resolved_fields_count > 0

    def drop_traces_for_field(self, key: str) -> None:
        """Removes traces for `key` so they are not displayed"""
        del self.traces[key]


def flatten_traces(
    nested_traces: NestedLookupTraces, remove_values: bool = True
) -> List[LookupTrace]:
    all_traces: List[LookupTrace] = []
    for trace in nested_traces:
        if isinstance(trace, LookupTraces):
            for field_trace in trace.traces.values():
                all_traces.extend(flatten_traces(field_trace, remove_values=remove_values))
        else:
            # drop all values as they may contain secrets
            if remove_values:
                trace = trace._replace(value=None)
            all_traces.append(trace)
    return all_traces


def get_run_context_warning(main_module_path: str) -> str:
    """Generates additional warnings when config resolution failed. There are two typical reasons
    (except genuine missing configurations):
    * `main_module_path` differs from current run_dir so `dlt` does not find related .dlt folder with settings
    * command line was used but pipeline script hardcodes config and secrets
    """
    from dlt.common.configuration.container import Container
    from dlt.common.runtime import run_context
    from dlt.common.pipeline import SupportsPipeline, PipelineContext

    msg = "\n"
    settings_moved_msg = (
        "If you keep `.dlt` folder with secret files in the same directory as your pipeline script"
        " but run your script or a dlt cli command from some other folder,"
        " secrets/configs will not be found.\n"
    )

    # find active pipeline
    context = Container()[PipelineContext]
    active_pipeline: SupportsPipeline = None
    # it is not always present
    if context.is_active():
        active_pipeline = context.pipeline()

    # use abs path to settings
    active_context = run_context.active()
    active_settings = os.path.abspath(active_context.settings_dir)

    if active_pipeline and active_pipeline.last_run_context:
        # warn if settings are now somewhere else than when pipeline had last successful run
        last_run_settings = active_pipeline.last_run_context["settings_dir"]
        if last_run_settings != active_settings:
            msg += (
                f"WARNING: Active pipeline `{active_pipeline.pipeline_name}` used"
                f" `{last_run_settings}` directory to read config and secrets for the last"
                f" successful run. Different directory `{active_settings}` is used now which may be"
                " the reason for configuration not resolving.\n"
                + settings_moved_msg
            )

        else:
            # locations didn't change
            pass
    else:
        # there's no active pipeline or it does not contain run context from successful run
        if main_module_path and main_module_path.endswith(".py"):
            pipeline_script_dir = os.path.abspath(os.path.dirname(main_module_path))
            run_dir = os.path.abspath(run_context.active().run_dir)
            if pipeline_script_dir != run_dir:
                msg += (
                    "WARNING: Your run dir (%s) is different from directory of your"
                    " pipeline script (%s).\n" % (run_dir, pipeline_script_dir)
                ) + settings_moved_msg
        else:
            # if no module path (ie. interactive) or it was not a script (ie. `dlt` cli`)
            pass

    if main_module_path and main_module_path.endswith("dlt"):
        msg += (
            "\nWARNING: When accessing data in the pipeline from the command line `dlt` will not"
            " execute user code and will just attach to the existing pipeline working dir. If you"
            " hardcoded your credentials or passed them explicitly ie. with `dlt.secrets` they"
            " won't be visible. Use environment variables, config files or other providers with"
            " config injection in order to use cli commands that access data.\n"
        )

    return msg


class UnmatchedConfigHintResolversException(ConfigurationException):
    """Raised when using `@resolve_type` on a field that doesn't exist in the spec"""

    def __init__(self, spec_name: str, field_names: Sequence[str]) -> None:
        self.field_names = field_names
        self.spec_name = spec_name
        example = f">>> class {spec_name}(BaseConfiguration)\n" + "\n".join(
            f">>>    {name}: Any" for name in field_names
        )
        msg = (
            f"The config spec `{spec_name}` has dynamic type resolvers for fields: `{field_names}`"
            " but these fields are not defined in the spec.\nWhen using @resolve_type() decorator,"
            f" Add the fields with 'Any' or another common type hint, example:\n\n{example}"
        )
        super().__init__(msg)


class FinalConfigFieldException(ConfigurationException):
    """rises when field was annotated as final ie Final[str] and the value is modified by config provider"""

    def __init__(self, spec_name: str, field: str) -> None:
        super().__init__(
            f"Field `{field}` in spec `{spec_name}` is final but is being changed by a config"
            " provider"
        )


class ConfigValueCannotBeCoercedException(ConfigurationValueError):
    """raises when value returned by config provider cannot be coerced to hinted type"""

    def __init__(self, field_name: str, field_value: Any, hint: type) -> None:
        self.field_name = field_name
        self.field_value = field_value
        self.hint = hint
        super().__init__(
            f"Configured value for field `{field_name}` cannot be coerced into type `{str(hint)}`"
        )


# class ConfigIntegrityException(ConfigurationException):
#     """thrown when value from ENV cannot be coerced to hinted type"""

#     def __init__(self, attr_name: str, env_value: str, info: Union[type, str]) -> None:
#         self.attr_name = attr_name
#         self.env_value = env_value
#         self.info = info
#         super().__init__('integrity error for field %s. %s.' % (attr_name, info))


class ConfigFileNotFoundException(ConfigurationException):
    """thrown when configuration file cannot be found in config folder"""

    def __init__(self, path: str) -> None:
        super().__init__(f"Missing config file in `{path}`")


class ConfigFieldMissingTypeHintException(ConfigurationException):
    """thrown when configuration specification does not have type hint"""

    def __init__(self, field_name: str, spec: Type[Any]) -> None:
        self.field_name = field_name
        self.typ_ = spec
        super().__init__(
            f"Field `{field_name}` on configspec `{spec}` does not provide required type hint"
        )


class ConfigFieldTypeHintNotSupported(ConfigurationException):
    """thrown when configuration specification uses not supported type in hint"""

    def __init__(self, field_name: str, spec: Type[Any], typ_: Type[Any]) -> None:
        self.field_name = field_name
        self.typ_ = spec
        super().__init__(
            f"Field `{field_name}` on configspec `{spec}` has hint with unsupported type `{typ_}`"
        )


class ValueNotSecretException(ConfigurationException):
    def __init__(self, provider_name: str, key: str) -> None:
        self.provider_name = provider_name
        self.key = key
        super().__init__(
            f"Provider `{provider_name}` cannot hold secret values but key `{key}` with secret"
            " value is present"
        )


class InvalidNativeValue(ConfigurationException):
    def __init__(
        self,
        spec: Type[Any],
        native_value_type: Type[Any],
        embedded_sections: Tuple[str, ...],
        inner_exception: Exception,
    ) -> None:
        self.spec = spec
        self.native_value_type = native_value_type
        self.embedded_sections = embedded_sections
        self.inner_exception = inner_exception
        inner_msg = f" {self.inner_exception}" if inner_exception is not ValueError else ""
        super().__init__(
            f"`{spec.__name__}` cannot parse the configuration value provided. The value is of type"
            f" `{native_value_type.__name__}` and comes from the sections `{embedded_sections}`"
            f" Value may be a secret and is not shown. Details: {inner_msg}"
        )


class ContainerInjectableContextMangled(ContainerException):
    def __init__(self, spec: Type[Any], existing_config: Any, expected_config: Any) -> None:
        self.spec = spec
        self.existing_config = existing_config
        self.expected_config = expected_config
        super().__init__(
            f"When restoring context `{spec.__name__}`, instance `{expected_config}` was expected,"
            f" instead instance `{existing_config}` was found."
        )


class ContextDefaultCannotBeCreated(ContainerException, KeyError):
    def __init__(self, spec: Type[Any]) -> None:
        self.spec = spec
        super().__init__(f"Container cannot create the default value of context `{spec.__name__}`.")


class DuplicateConfigProviderException(ConfigProviderException):
    def __init__(self, provider_name: str) -> None:
        super().__init__(
            provider_name,
            f"Provider with name `{provider_name}` already present in `ConfigProvidersContext`",
        )



try:
    import sqlalchemy  # noqa: I251
except ImportError:
    from dlt.common.libs.sql_alchemy_shims import URL, make_url
else:
    from sqlalchemy.engine import URL, make_url  # type: ignore[assignment]



"""Query LanceDB using DuckDB SQL client

LanceDB doesn't have an SQL interface. However, using a
DuckDB instance with the `lance-duckdb` extension allows
to read the `.lance` files (each maps to a single table).

This SQL client makes LanceDB compatible with the `dlt.Dataset`
inferface.
"""
from __future__ import annotations

from contextlib import contextmanager
from typing import Any, AnyStr, Iterator, TYPE_CHECKING

import sqlglot
import sqlglot.expressions as exp
import duckdb

import dlt
from dlt.destinations.exceptions import DatabaseUndefinedRelation
from dlt.common.destination.dataset import DBApiCursor
from dlt.common.destination.capabilities import DestinationCapabilitiesContext
from dlt.destinations.sql_client import raise_database_error, raise_open_connection_error
from dlt.destinations.impl.duckdb.sql_client import DuckDbSqlClient
from dlt.destinations.impl.duckdb.factory import _set_duckdb_raw_capabilities

if TYPE_CHECKING:
    from sqlglot import expressions as sge
    from duckdb import DuckDBPyConnection

    from dlt.destinations.impl.lancedb.lancedb_client import LanceDBClient


def _get_lancedb_sql_capabilities() -> DestinationCapabilitiesContext:
    caps = DestinationCapabilitiesContext()
    caps = _set_duckdb_raw_capabilities(caps)
    caps.preferred_loader_file_format = "parquet"
    return caps


def _install_and_load_lance_duckdb_extension(duckdb_con: DuckDBPyConnection) -> None:
    """Ensure the `lance-duckdb` extension is loaded.

    DuckDB ensures installation is only done once per system.
    Extension loading must be done on every connection
    """
    duckdb_con.execute("INSTALL lance FROM community;")
    duckdb_con.execute("LOAD lance;")


def _create_and_use_duckdb_dataset(
    duckdb_con: DuckDBPyConnection, dataset_qualified_name: str
) -> None:
    """Create a schema in the ephemeral DuckDB client that matches the `dlt` dataset name."""
    create_schema_sql = f"CREATE SCHEMA IF NOT EXISTS {dataset_qualified_name}"
    duckdb_con.execute(f"{create_schema_sql}; USE {dataset_qualified_name}")


def get_lance_table_uri(lancedb_client: LanceDBClient, table_name: str) -> str:
    """Create a URI for a Lance table

    This should be equivalent to
    ```python
    lancedb_client.credentials.get_conn().open_table("foo").to_lance().uri
    ```
    """
    dataset_lance_uri = lancedb_client.config.lance_uri
    qualified_table_name = lancedb_client.make_qualified_table_name(table_name)
    return f"{dataset_lance_uri}/{qualified_table_name}.lance"


def _prepare_create_view_statement(lance_table_uri: str, view_name: str) -> str:
    return f'CREATE OR REPLACE VIEW {view_name} AS SELECT * FROM "{lance_table_uri}"'


class LanceDBSQLClient(DuckDbSqlClient):
    def __init__(self, lancedb_client: LanceDBClient) -> None:
        self.lancedb_client = lancedb_client
        super().__init__(
            dataset_name=self.lancedb_client.dataset_name,
            staging_dataset_name=None,
            credentials=None,  # duckdb doesn't need special credentials
            capabilities=_get_lancedb_sql_capabilities(),
        )
        self._conn: DuckDBPyConnection | None = None

    @raise_open_connection_error
    def open_connection(self) -> DuckDBPyConnection:
        if self._conn:
            return self._conn

        self._conn = duckdb.connect(":memory:")
        _install_and_load_lance_duckdb_extension(self._conn)

        # by default, LanceDB has `dataset_name=None`. To be consistent, it uses DuckDB's
        # main schema by default
        if self.lancedb_client.dataset_name:
            _create_and_use_duckdb_dataset(self._conn, self.fully_qualified_dataset_name())

        return self._conn

    def close_connection(self) -> None:
        if self._conn:
            self._conn = None

    @contextmanager
    @raise_database_error
    def execute_query(
        self, query: str, *args: Any, **kwargs: Any  # type: ignore[override]
    ) -> Iterator[DBApiCursor]:
        # replace generic string placeholder by DuckDB placeholders
        if args or kwargs:
            query = query.replace("%s", "?")

        expression: sge.Expression = sqlglot.maybe_parse(query)
        for table in expression.find_all(exp.Table):
            if not table.this:
                continue

            self.create_view(table.name)

        with super().execute_query(query, *args, **kwargs) as cursor:
            yield cursor

    def create_view(self, table_name: str) -> None:
        lance_table_uri = get_lance_table_uri(self.lancedb_client, table_name)

        # lancedb allows omitting the dataset_name, calling `make_qualified_table_name` will
        # prepend the dataset_name even if it is None
        if self.dataset_name:
            view_name = self.make_qualified_table_name(table_name)
        else:
            view_name = self.capabilities.escape_identifier(
                self.capabilities.casefold_identifier(table_name)
            )

        create_view_sql = _prepare_create_view_statement(
            lance_table_uri=lance_table_uri,
            view_name=view_name,
        )
        try:
            self.open_connection().execute(create_view_sql)
        # Creating a DuckDB view will fail if the table doesn't exist in lance
        # potential edge case: a table only exists in the ephemeral DuckDB
        except duckdb.IOException as e:
            raise DatabaseUndefinedRelation(e)



import warnings

from dlt.common.warnings import Dlt100DeprecationWarning


def location_on_credentials_deprecated(arg: str) -> None:
    warnings.warn(
        f"Usage of `{arg}` option on Qdrant credentials is deprecated. Please set the `{arg}` on"
        f' Qdrant config. for example using toml section:\n[destination.qdrant]\n{arg}="value"\n',
        Dlt100DeprecationWarning,
        stacklevel=1,
    )



import contextlib
from copy import copy
from functools import partial
import inspect
from typing import (
    ClassVar,
    Dict,
    Iterable,
    Generator,
    List,
    Optional,
    Sequence,
    Tuple,
    Any,
)
from typing_extensions import Self

from dlt.common import logger
from dlt.common.configuration.resolve import inject_section
from dlt.common.configuration.specs import known_sections
from dlt.common.configuration.specs.base_configuration import ContainerInjectableContext, configspec
from dlt.common.configuration.specs.config_section_context import ConfigSectionContext
from dlt.common.normalizers.json.relational import DataItemNormalizer as RelationalNormalizer
from dlt.common.normalizers.json.typing import RelationalNormalizerConfig
from dlt.common.schema import Schema
from dlt.common.schema.typing import TColumnName, TSchemaContract
from dlt.common.schema.utils import normalize_table_identifiers
from dlt.common.typing import StrAny, TDataItem
from dlt.common.configuration.container import Container
from dlt.common.pipeline import (
    PipelineContext,
    StateInjectableContext,
    SupportsPipelineRun,
    pipeline_state,
)
from dlt.common.utils import (
    graph_find_scc_nodes,
    flatten_list_or_items,
    graph_edges_to_nodes,
    simple_repr,
    without_none,
)

from dlt.extract.items import SupportsPipe, TDecompositionStrategy
from dlt.extract.state import source_state
from dlt.extract.pipe_iterator import ManagedPipeIterator
from dlt.extract.pipe import Pipe
from dlt.extract.resource import DltResource
from dlt.extract.exceptions import (
    DataItemRequiredForDynamicTableHints,
    ResourceNotFoundError,
    ResourcesNotFoundError,
    DeletingResourcesNotSupported,
    InvalidParallelResourceDataType,
)


class DltResourceDict(Dict[str, DltResource]):
    def __init__(self, source_name: str, source_section: str) -> None:
        super().__init__()
        self.source_name = source_name
        self.source_section = source_section

        self._suppress_clone_on_setitem = False
        self._new_pipes: List[Pipe] = []
        """Collects pipes added via __setitem__"""
        self._cloned_pairs: Dict[str, Pipe] = {}
        """Maps pipe instance_id to pipe instance (contains full pipe tree)"""
        self._resource_pipes: Dict[str, DltResource] = {}
        """Maps pipe instance_id to resource (contains explicit and implicit resources)"""

    @property
    def selected(self) -> Dict[str, DltResource]:
        """Returns a subset of all resources that will be extracted and loaded to the destination."""
        return {k: v for k, v in self.items() if v.selected}

    # NOTE the name `extracted` suggests that the resources were ran and data was extracted
    # this is not what the docstring implies. Could rename to "`to_extract`". Or `selected` vs. `user_selected`
    @property
    def extracted(self) -> List[DltResource]:
        """Returns a all resources that will be extracted. That includes selected resources and all their parents.
        Note that several resources with the same name may be included if there were several parent instances of the
        same resource.
        """
        extracted = []
        for pipe in self.extracted_pipes:
            with contextlib.suppress(ResourceNotFoundError):
                extracted.append(self.with_pipe(pipe))
        return extracted

    @property
    def selected_dag(self) -> List[Tuple[str, str]]:
        """Returns a list of edges of directed acyclic graph of pipes and their parents in selected resources"""
        dag: List[Tuple[str, str]] = []
        for pipe in self.selected_pipes:
            selected = pipe
            parent: Pipe = None
            while (parent := pipe.parent) is not None:
                if not parent.is_empty:
                    dag.append((pipe.parent.name, pipe.name))
                    pipe = parent
                else:
                    # do not descend into disconnected pipes
                    break
            if selected is pipe:
                # NOTE a node pointing to itself creates a cycle / is not a DAG
                # the implementation doesn't match the function name and docstring
                # add isolated element
                # TODO: it is not node pointing to itself. change to (pipe.name, None)
                #   to indicate isolated node
                dag.append((pipe.name, pipe.name))
        return dag

    @property
    def pipes(self) -> List[Pipe]:
        return [r._pipe for r in self.values()]

    @property
    def selected_pipes(self) -> Sequence[Pipe]:
        return [r._pipe for r in self.values() if r.selected]

    @property
    def extracted_pipes(self) -> Sequence[Pipe]:
        """Returns all execution pipes in the source including all parent pipes"""
        extracted = []
        for resource in self.selected.values():
            pipe = resource._pipe
            while pipe is not None:
                extracted.append(pipe)
                pipe = pipe.parent
        return extracted

    def select(self, *resource_names: str) -> Dict[str, DltResource]:
        """Selects `resource_name` to be extracted, and unselects remaining resources."""
        for name in resource_names:
            if name not in self:
                # if any key is missing, display the full info
                raise ResourcesNotFoundError(
                    self.source_name, set(self.keys()), set(resource_names)
                )
        # set the selected flags
        for resource in self.values():
            self[resource.name].selected = resource.name in resource_names
        return self.selected

    def with_pipe(self, pipe: SupportsPipe) -> DltResource:
        """Gets resource with given execution pipe matched by pipe instance id.
        Note that we do not use names to match pipe to resource as many resources
        with the same name may be extracted at the same time: ie. when transformers
        with different names depend on several instances of the same resource.
        Note that ad hoc resources for parent pipes without resource won't be created
        """
        try:
            return self._resource_pipes[pipe.instance_id]
        except KeyError:
            assert pipe.name not in self, (
                f"consistency of search by pipe id failed, {pipe.name} was found,"
                f" {pipe.instance_id} not."
            )
            raise ResourceNotFoundError(
                pipe.name,
                f"Resource with pipe with `{pipe.instance_id} could not be found in source",
            )

    def add(self, *resources: DltResource) -> None:
        """Add `resources` to the source. Adding multiple resources with the same name is not
        supported.
        """
        with self._add_multiple_resources():
            for resource in resources:
                if resource.name in self:
                    # for resources with the same name try to add the resource as an another pipe
                    self[resource.name].add_pipe(resource)
                else:
                    self[resource.name] = resource

        self._clone_new_pipes([r.name for r in resources])
        self._adjust_instances()

    def detach(self, resource_name: str = None) -> DltResource:
        """Clones `resource_name` (including parent resource pipes) and removes source contexts.
        Defaults to the first resource in the source if `resource_name` is None.
        """
        return (self[resource_name] if resource_name else list(self.values())[0])._clone(
            with_parent=True
        )

    @contextlib.contextmanager
    def _add_multiple_resources(self) -> Generator[TDataItem, None, None]:
        # temporarily block cloning when single resource is added
        try:
            self._suppress_clone_on_setitem = True
            yield
        finally:
            self._suppress_clone_on_setitem = False

    def _clone_new_pipes(self, resource_names: Sequence[str]) -> None:
        # get tree of new pipes and updated
        _, self._cloned_pairs = ManagedPipeIterator.clone_pipes(self._new_pipes, self._cloned_pairs)
        # replace pipes in resources, the cloned_pipes preserve parent connections
        for name in resource_names:
            resource = self[name]
            pipe_id = resource._pipe.instance_id
            resource._pipe = self._cloned_pairs[pipe_id]
            # store pipes of newly added resources
            self._resource_pipes[pipe_id] = resource

        self._new_pipes.clear()

    def _adjust_instances(self) -> None:
        """Walks resource tree starting from explicit resources and adjusts parent resource instances
        and pipe instances based on pipe instance_id. This is necessary to keep consistent
        pipe and resource tree. Also detects implicitly added parents and adds those resources
        to resource_pipes.
        Dangling resources and pipes are removed at the end
        """
        # collect all pipes and resources
        collected_pipes = set()
        collected_resources = set()
        # descend into parents and align _parent instances
        for resource in self.values():
            parent = resource._parent
            assert resource._pipe is self._cloned_pairs[resource._pipe.instance_id]
            collected_pipes.add(resource._pipe)
            collected_resources.add(resource)
            while parent and not parent._pipe.is_empty:
                # also index pipes of parents that are not part of the source
                if parent._pipe.instance_id not in self._resource_pipes:
                    # clone parent when adding
                    parent = copy(parent)
                    # correct pipe
                    self._resource_pipes[parent._pipe.instance_id] = parent
                # correct resource parent
                parent = self.with_pipe(resource._pipe.parent)
                parent._pipe = self._cloned_pairs[parent._pipe.instance_id]
                resource._parent = parent
                # recur
                collected_pipes.add(parent._pipe)
                collected_resources.add(parent)
                resource = parent
                parent = parent._parent

        # drop all resource instances and pipe instances that were not collected in tree walk
        # those are no longer used and will not be extracted. note that both explicitly
        # added and implicitly (ie. transformer parent) are collected
        dropped_resources = set(self._resource_pipes.values()).difference(collected_resources)
        if dropped_resources:
            # all collected resources must be known
            assert dropped_resources == set(self._resource_pipes.values()).symmetric_difference(
                collected_resources
            )
            for resource in list(self._resource_pipes.values()):
                if resource in dropped_resources:
                    del self._resource_pipes[resource._pipe.instance_id]
        dropped_pipes = set(self._cloned_pairs.values()).difference(collected_pipes)
        if dropped_pipes:
            # all collected pipes must be known
            assert dropped_pipes == set(self._cloned_pairs.values()).symmetric_difference(
                collected_pipes
            )
            for pipe in list(self._cloned_pairs.values()):
                if pipe in dropped_pipes:
                    del self._cloned_pairs[pipe.instance_id]

        self._assert_source_consistency()

    def _assert_source_consistency(self) -> None:
        """Makes sure that all resource and pipe references in extracted resource tree
        point to the same Python instances.
        """
        pipe_ids = list(self._cloned_pairs.values())
        for resource in self.values():
            if not resource._pipe.is_empty:
                assert resource._pipe in pipe_ids
                assert resource is self.with_pipe(resource._pipe)
                # check parent
                parent = resource._parent
                while parent:
                    # also pipe has parent
                    assert resource._pipe.has_parent
                    # instances must match if known
                    if parent.name in self:
                        if parent is not self[parent.name]:
                            logger.warning(
                                f"Source `{self.source_name}` contains resource `{resource.name}`"
                                f" which has a parent named `{parent.name}` which refers to a"
                                " different instance than an explicitly added resource with the"
                                " same name. This may happen if you added resource and transformer"
                                " to your source and later replaced the resource (parent) with"
                                f" another resource. In that case transformer `{resource.name}`"
                                " still points to original parent which will be separately"
                                " extracted. If this is not what you want, replace transformer as"
                                " well or rename your resources and transformers."
                            )

                    if parent._pipe.is_empty:
                        assert parent._pipe not in pipe_ids
                    else:
                        # knows pipes
                        assert parent._pipe in pipe_ids
                        assert parent is self.with_pipe(parent._pipe)
                        # pipes must match
                        assert parent._pipe is resource._pipe.parent

                    resource = parent
                    parent = resource._parent

    def __setitem__(self, resource_name: str, resource: DltResource) -> None:
        if resource_name != resource.name:
            raise ValueError(
                f"The index name `{resource_name}` does not correspond to resource name"
                f" `{resource.name}`"
            )
        pipe_instance_id = resource._pipe.instance_id
        # make shallow copy of the resource
        resource = copy(resource)
        # resource.section = self.source_section
        resource.source_name = self.source_name
        if pipe_instance_id not in self._cloned_pairs:
            self._new_pipes.append(resource._pipe)
        # now set it in dict
        super().__setitem__(resource_name, resource)

        # immediately clone pipe if not suppressed
        if not self._suppress_clone_on_setitem:
            self._clone_new_pipes([resource.name])
            self._adjust_instances()

    def __delitem__(self, resource_name: str) -> None:
        raise DeletingResourcesNotSupported(self.source_name, resource_name)

    def __repr__(self) -> str:
        if not self:
            return "{}"

        s = "{\n  'selected': {"
        for name, r in self.selected.items():
            s += f"\n    '{name}': {r.__repr__()}, "
        s += "\n  }"

        s += "\n  'extracted': {"
        for r in self.extracted:
            s += f"\n    '{r.name}': {r.__repr__()}, "
        s += "\n  }"
        s += "\n}"

        return s


class DltSource(Iterable[TDataItem]):
    """Groups several `dlt resources` under a single schema and allows to perform operations on them.

    The instance of this class is created whenever you call the `dlt.source` decorated function. It automates several functions for you:
    * You can pass this instance to `dlt` `run` method in order to load all data present in the `dlt resources`.
    * You can select and deselect resources that you want to load via `with_resources` method
    * You can access the resources (which are `DltResource` instances) as source attributes
    * It implements `Iterable` interface so you can get all the data from the resources yourself and without dlt pipeline present.
    * It will create a DAG from resources and transformers and optimize the extraction so parent resources are extracted only once
    * You can get the `schema` for the source and all the resources within it.
    * You can use a `run` method to load the data with a default instance of dlt pipeline.
    * You can get source read only state for the currently active Pipeline instance
    """

    def __init__(
        self, schema: Schema, section: str, resources: Sequence[DltResource] = None
    ) -> None:
        self.section = section
        """Tells if iterator associated with a source is exhausted"""
        self._schema = schema
        self._resources: DltResourceDict = DltResourceDict(self.name, self.section)

        if resources:
            self.resources.add(*resources)

    @classmethod
    def from_data(cls, schema: Schema, section: str, data: Any) -> Self:
        """Converts any `data` supported by `dlt` `run` method into `dlt source` with a name `section`.`name` and `schema` schema."""
        # creates source from various forms of data
        if isinstance(data, DltSource):
            return data  # type: ignore[return-value]

        # in case of sequence, enumerate items and convert them into resources
        if isinstance(data, Sequence):
            resources = [DltResource.from_data(i) for i in data]
        else:
            resources = [DltResource.from_data(data)]

        return cls(schema, section, resources)

    @property
    def name(self) -> str:
        return self._schema.name

    @property
    def max_table_nesting(self) -> int:
        """A schema hint that sets the maximum depth of nested table above which the remaining nodes are loaded as structs or JSON."""
        return RelationalNormalizer.get_normalizer_config(self._schema).get("max_nesting")

    @max_table_nesting.setter
    def max_table_nesting(self, value: int) -> None:
        if value is None:
            # this also check the normalizer type
            config = RelationalNormalizer.get_normalizer_config(self._schema)
            config.pop("max_nesting", None)
        else:
            RelationalNormalizer.update_normalizer_config(self._schema, {"max_nesting": value})

    @property
    def root_key(self) -> Optional[bool]:
        """Enables merging on all resources by propagating root foreign key to nested tables.
        This option is most useful if you plan to change write disposition of a resource to disable/enable merge.

        """
        # this also check the normalizer type
        config = RelationalNormalizer.get_normalizer_config(self._schema)
        is_root_key = config.get("root_key_propagation")
        if is_root_key is None:
            # if not found get legacy value
            is_root_key = self._get_root_key_legacy(config)
            if is_root_key:
                # set the root key if legacy value set
                self.root_key = True
        return is_root_key

    @root_key.setter
    def root_key(self, value: bool) -> None:
        # this also check the normalizer type
        config = RelationalNormalizer.get_normalizer_config(self._schema)
        if value is None:
            value = self._get_root_key_legacy(config)
        if value is not None:
            RelationalNormalizer.update_normalizer_config(
                self._schema,
                {"root_key_propagation": value},
            )

    def _get_root_key_legacy(self, config: RelationalNormalizerConfig) -> Optional[bool]:
        data_normalizer = self._schema.data_item_normalizer
        assert isinstance(data_normalizer, RelationalNormalizer)
        # we must remove root key propagation
        with contextlib.suppress(KeyError):
            propagation_config = config["propagation"]
            propagation_config["root"].pop(data_normalizer.c_dlt_id)
            # and set the value below
            return True
        return None

    @property
    def schema_contract(self) -> TSchemaContract:
        return self.schema.settings.get("schema_contract")

    @schema_contract.setter
    def schema_contract(self, settings: TSchemaContract) -> None:
        self.schema.set_schema_contract(settings)

    @property
    def exhausted(self) -> bool:
        """Check all selected pipes whether one of them has started. if so, the source is exhausted."""
        for pipe in self._resources.selected_pipes:
            item = pipe.gen
            if inspect.isgenerator(item):
                if inspect.getgeneratorstate(item) != "GEN_CREATED":
                    return True
        return False

    @property
    def resources(self) -> DltResourceDict:
        """A dictionary of all resources present in the source, where the key is a resource name.

        Returns:
            DltResourceDict: A dictionary of all resources present in the source, where the key is a resource name.
        """
        return self._resources

    @property
    def selected_resources(self) -> Dict[str, DltResource]:
        """A dictionary of all the resources that are selected to be loaded.

        Returns:
            Dict[str, DltResource]: A dictionary of all the resources that are selected to be loaded.
        """
        return self._resources.selected

    @property
    def schema(self) -> Schema:
        return self._schema

    @schema.setter
    def schema(self, value: Schema) -> None:
        self._schema = value

    def discover_schema(self, item: TDataItem = None, meta: Any = None) -> Schema:
        """Computes table schemas for all selected resources in the source and merges them with a copy of current source schema. If `item` is provided,
        dynamic tables will be evaluated, otherwise those tables will be ignored."""
        schema = self._schema.clone(update_normalizers=True)
        for r in self.selected_resources.values():
            # names must be normalized here
            with contextlib.suppress(DataItemRequiredForDynamicTableHints):
                root_table_schema = r.compute_table_schema(item, meta)
                nested_tables_schema = r.compute_nested_table_schemas(
                    root_table_schema["name"], schema.naming, item, meta
                )
                # NOTE must ensure that `schema.update_table()` is called in an order that respect parent-child relationships
                for table_schema in (root_table_schema, *nested_tables_schema):
                    partial_table = normalize_table_identifiers(table_schema, self._schema.naming)
                    schema.update_table(partial_table)
        return schema

    def with_resources(self, *resource_names: str) -> "DltSource":
        """A convenience method to select one of more resources to be loaded. Returns a clone of the original source with the specified resources selected."""
        source = self.clone()
        source._resources.select(*resource_names)
        return source

    def decompose(self, strategy: TDecompositionStrategy) -> List["DltSource"]:
        """Decomposes source into a list of sources with a given strategy.

        "none" will return source as is
        "scc" will decompose the dag of selected pipes and their parent into strongly connected components
        """
        if strategy == "none":
            return [self]
        elif strategy == "scc":
            dag = self.resources.selected_dag
            scc = graph_find_scc_nodes(graph_edges_to_nodes(dag, directed=False))
            # components contain elements that are not currently selected
            selected_set = set(self.resources.selected.keys())
            return [self.with_resources(*component.intersection(selected_set)) for component in scc]
        else:
            raise ValueError(strategy)

    def add_limit(
        self,
        max_items: Optional[int] = None,
        max_time: Optional[float] = None,
        count_rows: Optional[bool] = False,
    ) -> "DltSource":  # noqa: A003
        """Limits the items processed in all selected resources in the source that are not transformers: by count or time.

        This is useful for testing, debugging and generating sample datasets for experimentation. You can easily get your test dataset in a few minutes, when otherwise
        you'd need to wait hours for the full loading to complete.

        For incremental resources that return rows in deterministic order, you can use this function to process large load
        in batches.

        Notes:
            1. Transformers resources won't be limited. They should process all the data they receive fully to avoid inconsistencies in generated datasets.
            2. Each yielded item may contain several records. `add_limit` only limits the "number of yields", not the total number of records.
            3. Empty pages/yields are also counted. Use `count_rows` to skip empty pages.

        Args:
            max_items (Optional[int]): The maximum number of items (not rows!) to yield, set to None for no limit
            max_time (Optional[float]): The maximum number of seconds for this generator to run after it was opened, set to None for no limit
            count_rows (Optional[bool]): Default: false.
                Count rows instead of pages. Note that if resource yields pages of rows, last page will not be trimmed and more rows that expected will be received.
        Returns:
            "DltSource": returns self
        """
        for resource in self.resources.selected.values():
            resource.add_limit(max_items, max_time=max_time, count_rows=count_rows)
        return self

    def parallelize(self) -> "DltSource":
        """Mark all resources in the source to run in parallel.

        Only transformers and resources based on generators and generator functions are supported, unsupported resources will be skipped.
        """
        for resource in self.resources.selected.values():
            try:
                resource.parallelize()
            except InvalidParallelResourceDataType:
                pass
        return self

    @property
    def run(self) -> SupportsPipelineRun:
        """A convenience method that will call `run` run on the currently active `dlt` pipeline. If pipeline instance is not found, one with default settings will be created."""
        self_run: SupportsPipelineRun = partial(
            Container()[PipelineContext].pipeline().run, data=self
        )
        return self_run

    @property
    def state(self) -> StrAny:
        """Gets source-scoped state from the active pipeline. PipelineStateNotAvailable is raised if no pipeline is active"""
        with inject_section(self._get_config_section_context()):
            return source_state()

    def clone(self, with_name: str = None) -> "DltSource":
        """Creates a deep copy of the source where copies of schema, resources and pipes are created.

        If `with_name` is provided, a schema is cloned with a changed name
        """
        # mind that resources and pipes are cloned when added to the DltResourcesDict in the source constructor
        return DltSource(
            self.schema.clone(with_name=with_name), self.section, list(self._resources.values())
        )

    def __iter__(self) -> Generator[TDataItem, None, None]:
        """Opens iterator that yields the data items from all the resources within the source in the same order as in Pipeline class.

        A read-only state is provided, initialized from active pipeline state. The state is discarded after the iterator is closed.

        A source config section is injected to allow secrets/config injection as during regular extraction.
        """
        # use the same state dict when opening iterator and when iterator is iterated
        mock_state, _ = pipeline_state(Container(), {})
        state_context = StateInjectableContext(state=mock_state)
        section_context = self._get_config_section_context()
        source_context = SourceInjectableContext(self)
        schema_context = SourceSchemaInjectableContext(self.schema)

        # managed pipe iterator will set the context on each call to  __next__
        with (
            inject_section(section_context),
            Container().injectable_context(state_context),
            Container().injectable_context(source_context),
            Container().injectable_context(schema_context),
        ):
            pipe_iterator: ManagedPipeIterator = ManagedPipeIterator.from_pipes(self._resources.selected_pipes)  # type: ignore
        pipe_iterator.set_context([section_context, state_context, schema_context, source_context])
        _iter = map(lambda item: item.item, pipe_iterator)
        return flatten_list_or_items(_iter)

    def _get_config_section_context(self) -> ConfigSectionContext:
        proxy = Container()[PipelineContext]
        pipeline_name = None if not proxy.is_active() else proxy.pipeline().pipeline_name
        return ConfigSectionContext(
            pipeline_name=pipeline_name,
            sections=(known_sections.SOURCES, self.section, self.name),
            source_state_key=self.name,
        )

    def __getattr__(self, resource_name: str) -> DltResource:
        try:
            return self._resources[resource_name]
        except KeyError:
            all_resources = ", ".join(self._resources.keys())
            raise AttributeError(
                f"Resource `{resource_name}` not found in source `{self.name}`. Available"
                f" resources: {all_resources}"
            )

    def __setattr__(self, name: str, value: Any) -> None:
        if isinstance(value, DltResource):
            self.resources[name] = value
        else:
            super().__setattr__(name, value)

    def __repr__(self) -> str:
        kwargs = {
            "name": self.name,
            "section": self.section,
            "schema_contract": "{...}" if self.schema_contract else None,
            "exhausted": self.exhausted if self.exhausted else None,
            "n_resources": len(self.resources),
            "resources": list(self.resources.keys()),
        }
        return simple_repr("@dlt.source", **without_none(kwargs))

    def __str__(self) -> str:
        info = (
            f"DltSource `{self.name}` section `{self.section}` contains"
            f" {len(self.resources)} resource(s) of which {len(self.selected_resources)} are"
            " selected"
        )
        for r in self.resources.values():
            selected_info = "selected" if r.selected else "not selected"
            if r.is_transformer:
                info += (
                    f"\ntransformer `{r.name}` is {selected_info} and takes data from"
                    f" `{r._pipe.parent.name}`"
                )
            else:
                info += f"\nresource `{r.name}` is {selected_info}"
        if self.exhausted:
            info += (
                "\nSource is already iterated and cannot be used again ie. to display or load data."
            )
        else:
            info += (
                "\nIf you want to see the data items in this source you must iterate it or convert"
                " to list ie. `list(source)`."
            )
        info += " Note that, like any iterator, you can iterate the source only once."
        info += f"\ninstance id: {id(self)}"
        return info


@configspec
class SourceSchemaInjectableContext(ContainerInjectableContext):
    """A context containing the source schema, present when dlt.source/resource decorated function is executed"""

    schema: Schema = None

    can_create_default: ClassVar[bool] = False


@configspec
class SourceInjectableContext(ContainerInjectableContext):
    """A context containing the source schema, present when dlt.resource decorated function is executed"""

    source: DltSource = None

    can_create_default: ClassVar[bool] = False


class _DltSingleSource(DltSource):
    """Used to register standalone (non-inner) resources"""

    @property
    def single_resource(self) -> DltResource:
        return list(self.resources.values())[0]



from contextlib import contextmanager
import re
import threading
from typing import ClassVar, Dict, Iterator, Optional, Tuple, Type, TypeVar

from dlt.common.configuration.specs.base_configuration import (
    ContainerInjectableContext,
    TInjectableContext,
)
from dlt.common.configuration.exceptions import (
    ContainerInjectableContextMangled,
    ContextDefaultCannotBeCreated,
)
from dlt.common.typing import is_subclass


class Container:
    """A singleton injection container holding several injection contexts. Implements basic dictionary interface.

    Injection context is identified by its type and available via dict indexer. The common pattern is to instantiate default context value
    if it is not yet present in container.

    By default, the context is thread-affine so it can be injected only n the thread that originally set it. This behavior may be changed
    in particular context type (spec).

    The indexer is settable and allows to explicitly set the value. This is required by in any context that needs to be explicitly instantiated.

    The `injectable_context` allows to set a context with a `with` keyword and then restore the previous one after it gets out of scope.

    """

    _INSTANCE: ClassVar[Optional["Container"]] = None
    _LOCK: ClassVar[threading.RLock] = threading.RLock()
    _MAIN_THREAD_ID: ClassVar[int] = threading.get_ident()
    """A main thread id to which get item will fallback for contexts without default"""

    _TContext = Dict[Type[ContainerInjectableContext], ContainerInjectableContext]

    thread_contexts: Dict[int, Tuple["Container._TContext", threading.RLock]]
    """Thread-id → (context dict, per-context RLock). Every thread gets its own lock."""

    def __new__(cls: Type["Container"]) -> "Container":
        if not cls._INSTANCE:
            cls._INSTANCE = super().__new__(cls)
            cls._INSTANCE.thread_contexts = {Container._MAIN_THREAD_ID: ({}, threading.RLock())}

        return cls._INSTANCE

    def __init__(self) -> None:
        pass

    def __getitem__(self, spec: Type[TInjectableContext]) -> TInjectableContext:
        # return existing config object or create it from spec
        if not is_subclass(spec, ContainerInjectableContext):
            raise KeyError(f"`{spec.__name__}` is not a context")

        context, lock = self._thread_context(spec)
        item = context.get(spec)
        if item is None:
            if spec.can_create_default:
                with lock:
                    # re-check after acquiring lock (double-checked locking)
                    item = context.get(spec)
                    if item is None:
                        # create and set inside lock - RLock allows reentrant access
                        # if spec() or callbacks access container
                        item = spec()
                        self._thread_setitem(context, spec, item)
            else:
                raise ContextDefaultCannotBeCreated(spec)

        return item  # type: ignore[return-value]

    def __setitem__(self, spec: Type[TInjectableContext], value: TInjectableContext) -> None:
        # value passed to container must be final
        value.resolve()
        context, lock = self._thread_context(spec)
        with lock:
            self._thread_setitem(context, spec, value)

    def __delitem__(self, spec: Type[TInjectableContext]) -> None:
        context, lock = self._thread_context(spec)
        with lock:
            self._thread_delitem(context, spec)

    def __contains__(self, spec: Type[TInjectableContext]) -> bool:
        context, _ = self._thread_context(spec)
        return spec in context

    def _effective_thread_id(self) -> int:
        """Return the effective thread id for context lookup.

        Thread in a dlt pool (named `dlt-pool-{id}-*`) see the same context
        to support sharing context in isolated extract pipelines.
        """
        if m := re.match(r"dlt-pool-(\d+)-", threading.current_thread().name):
            return int(m.group(1))
        return threading.get_ident()

    def _thread_context(
        self, spec: Type[TInjectableContext]
    ) -> Tuple["Container._TContext", threading.RLock]:
        """Return (context dict, per-context lock) for the given spec."""
        if spec.global_affinity:
            return self.thread_contexts[Container._MAIN_THREAD_ID]

        thread_id = self._effective_thread_id()

        # fast path: context already exists (dict.get is GIL-atomic)
        ctx_and_lock = self.thread_contexts.get(thread_id)
        if ctx_and_lock is not None:
            return ctx_and_lock

        # create new thread context under structural lock
        with Container._LOCK:
            if (ctx_and_lock := self.thread_contexts.get(thread_id)) is None:
                ctx_and_lock = ({}, threading.RLock())
                self.thread_contexts[thread_id] = ctx_and_lock
            return ctx_and_lock

    def _thread_setitem(
        self,
        context: "Container._TContext",
        spec: Type[ContainerInjectableContext],
        value: TInjectableContext,
    ) -> None:
        """Set context item and run callbacks. Caller must hold the context lock."""
        old_ctx = context.get(spec)
        if old_ctx:
            old_ctx.before_remove()
            old_ctx.in_container = False
        context[spec] = value
        value.in_container = True
        value.after_add()
        if not value.extras_added:
            value.add_extras()
            value.extras_added = True

    def _thread_delitem(
        self,
        context: "Container._TContext",
        spec: Type[ContainerInjectableContext],
    ) -> None:
        """Delete context item and run callbacks. Caller must hold the context lock."""
        old_ctx = context.get(spec)
        if old_ctx is None:
            raise KeyError(spec)
        old_ctx.before_remove()
        old_ctx.in_container = False
        del context[spec]

    @contextmanager
    def injectable_context(
        self, config: TInjectableContext, lock_context_on_yield: bool = False
    ) -> Iterator[TInjectableContext]:
        """Insert ``config`` into the container and restore the previous value on exit.

        Args:
            config (TInjectableContext): The context instance to inject.
            lock_context_on_yield (bool): When False (default), the per-context lock
                is released before yield and re-acquired for restore, allowing other
                threads to write to the same spec. When True, the lock is held for
                the entire lifetime including yield.

        Yields:
            TInjectableContext: The injected ``config`` instance.

        Raises:
            ContainerInjectableContextMangled: If another writer overwrote the
                context during yield (only possible when ``lock_context_on_yield``
                is False).
        """

        config.resolve()
        spec = type(config)
        context, ctx_lock = self._thread_context(spec)

        # always lock for setitem
        ctx_lock.acquire()
        try:
            previous_config: ContainerInjectableContext = context.get(spec)
            self._thread_setitem(context, spec, config)
        except BaseException:
            ctx_lock.release()
            raise

        if not lock_context_on_yield:
            ctx_lock.release()

        try:
            yield config
        finally:
            # re-acquire for restore when lock_context is False
            if not lock_context_on_yield:
                ctx_lock.acquire()
            try:
                current_config = context.get(spec)
                if current_config is config:
                    if previous_config is None:
                        self._thread_delitem(context, spec)
                    else:
                        self._thread_setitem(context, spec, previous_config)
                else:
                    raise ContainerInjectableContextMangled(spec, context[spec], config)
            finally:
                ctx_lock.release()

    def get(self, spec: Type[TInjectableContext]) -> Optional[TInjectableContext]:
        try:
            return self[spec]
        except KeyError:
            return None

    def get_worker_contexts(
        self,
    ) -> Dict[Type[ContainerInjectableContext], ContainerInjectableContext]:
        """Collect all contexts marked with `worker_affinity=True` for the current thread.

        Returns a dict mapping context type to instance for all contexts that should
        be passed to worker processes.
        """
        result: Dict[Type[ContainerInjectableContext], ContainerInjectableContext] = {}

        with Container._LOCK:
            # collect global_affinity contexts from main thread
            main_ctx, _ = self.thread_contexts[Container._MAIN_THREAD_ID]
            for spec, ctx in list(main_ctx.items()):
                if spec.worker_affinity:
                    result[spec] = ctx

            # collect thread-local contexts from the effective thread's context
            thread_id = self._effective_thread_id()
            if thread_id != Container._MAIN_THREAD_ID:
                ctx_and_lock = self.thread_contexts.get(thread_id)
                if ctx_and_lock is not None:
                    thread_ctx, _ = ctx_and_lock
                    for spec, ctx in list(thread_ctx.items()):
                        if spec.worker_affinity:
                            result[spec] = ctx

        return result

    @staticmethod
    def thread_pool_prefix() -> str:
        """Creates a container friendly pool prefix that contains starting thread id. Container implementation will automatically use it
        for any thread-affine contexts instead of using id of the pool thread
        """
        return f"dlt-pool-{threading.get_ident()}-"



from contextlib import contextmanager
import sys
import os
import shutil
import types
import subprocess
from typing import Any, ClassVar, Generator, List, Type, ContextManager

from dlt.common import known_env
from dlt.common.exceptions import CannotInstallDependencies, VenvNotFound


class Venv:
    """Creates and wraps the Python Virtual Environment to allow for code execution"""

    PIP_TOOL: ClassVar[str] = os.environ.get(known_env.DLT_PIP_TOOL, None)

    def __init__(self, context: types.SimpleNamespace, current: bool = False) -> None:
        """Please use `Venv.create`, `Venv.restore` or `Venv.restore_current` methods to create Venv instance"""
        self.context = context
        self.current = current

    @classmethod
    def create(cls, venv_dir: str, dependencies: List[str] = None) -> "Venv":
        """Creates a new Virtual Environment at the location specified in `venv_dir` and installs `dependencies` via pip. Deletes partially created environment on failure."""
        import venv

        class DLTEnvBuilder(venv.EnvBuilder):
            context: types.SimpleNamespace

            def __init__(self) -> None:
                # setup pip only if autodetect or explicit pip, skip for uv etc.
                super().__init__(with_pip=Venv.get_pip_tool() == "pip", clear=True)

            def post_setup(self, context: types.SimpleNamespace) -> None:
                self.context = context

        b = DLTEnvBuilder()
        try:
            b.create(os.path.abspath(venv_dir))
            if dependencies:
                Venv.install_deps(b.context, dependencies)
        except Exception:
            if os.path.isdir(venv_dir):
                shutil.rmtree(venv_dir)
            raise
        return cls(b.context)

    @classmethod
    def restore(cls, venv_dir: str, current: bool = False) -> "Venv":
        """Restores Virtual Environment at `venv_dir`"""
        import venv

        if not os.path.isdir(venv_dir):
            raise VenvNotFound(venv_dir)
        b = venv.EnvBuilder(clear=False, upgrade=False)
        c = b.ensure_directories(os.path.abspath(venv_dir))
        if not os.path.isfile(c.env_exe):
            raise VenvNotFound(c.env_exe)
        return cls(c, current)

    @classmethod
    def restore_current(cls) -> "Venv":
        """Wraps the current Python environment."""
        try:
            venv = cls.restore(os.environ["VIRTUAL_ENV"], current=True)
        except KeyError:
            import sys

            # do not set bin path because it is not known
            context = types.SimpleNamespace(bin_path="", env_exe=sys.executable)
            venv = cls(context, current=True)
        return venv

    def __enter__(self) -> "Venv":
        if self.current:
            raise NotImplementedError("Context manager does not work with current venv")
        return self

    def __exit__(
        self, exc_type: Type[BaseException], exc_val: BaseException, exc_tb: types.TracebackType
    ) -> None:
        self.delete_environment()

    def delete_environment(self) -> None:
        """Deletes the Virtual Environment."""
        if self.current:
            raise NotImplementedError("Context manager does not work with current venv")
        # delete venv
        if self.context.env_dir and os.path.isdir(self.context.env_dir):
            shutil.rmtree(self.context.env_dir)

    def start_command(
        self, entry_point: str, *script_args: Any, **popen_kwargs: Any
    ) -> "subprocess.Popen[str]":
        command = os.path.join(self.context.bin_path, entry_point)
        cmd = [command, *script_args]
        return subprocess.Popen(cmd, **popen_kwargs)

    def run_command(self, entry_point: str, *script_args: Any) -> str:
        """Runs any `command` with specified `script_args`. Current `os.environ` and cwd is passed to executed process"""
        # runs one of installed entry points typically CLIs coming with packages and installed into PATH
        # command = os.path.join(self.context.bin_path, entry_point)
        command: str = None
        if os.path.basename(entry_point) == entry_point:
            # look into current venv
            command = shutil.which(entry_point, path=self.context.bin_path)

        if not command:
            # for all paths do a normal lookup
            command = entry_point

        cmd = [command, *script_args]
        return subprocess.check_output(
            cmd, stderr=subprocess.STDOUT, text=True, errors="backslashreplace"
        )

    def run_script(self, script_path: str, *script_args: Any) -> str:
        """Runs a python `script` source with specified `script_args`. Current `os.environ` and cwd is passed to executed process"""
        # os.environ is passed to executed process
        cmd = [self.context.env_exe, os.path.abspath(script_path), *script_args]
        try:
            return subprocess.check_output(
                cmd, stderr=subprocess.STDOUT, text=True, errors="backslashreplace"
            )
        except subprocess.CalledProcessError as cpe:
            if cpe.returncode == 2:
                raise FileNotFoundError(script_path)
            else:
                raise

    def run_module(self, module: str, *module_args: Any) -> str:
        """Runs a python `module` with specified `module_args`. Current `os.environ` and cwd is passed to executed process"""
        cmd = [self.context.env_exe, "-m", module, *module_args]
        return subprocess.check_output(
            cmd, stderr=subprocess.STDOUT, text=True, errors="backslashreplace"
        )

    def add_dependencies(self, dependencies: List[str] = None) -> None:
        Venv.install_deps(self.context, dependencies)

    @staticmethod
    def set_pip_tool(pip_tool: str) -> ContextManager[None]:
        """Sets pip tool in context manager, not thread safe"""

        @contextmanager
        def _ctx(pip_tool: str) -> Generator[None, None, None]:
            old_tool = Venv.PIP_TOOL
            try:
                Venv.PIP_TOOL = pip_tool
                yield
            finally:
                Venv.PIP_TOOL = old_tool

        return _ctx(pip_tool)

    @staticmethod
    def get_pip_tool() -> str:
        """Gets configured tool to manage python packages, using DLT_PIP_TOOL or by
        autodetecting `uv` and `pip`.
        """
        if Venv.PIP_TOOL is None:
            # autodetect tool
            import shutil

            return "uv" if shutil.which("uv") else "pip"
        return Venv.PIP_TOOL

    @staticmethod
    def get_pip_command(context: types.SimpleNamespace, command: str) -> List[str]:
        """Returns a sequence that is a pip-like command"""
        pip_tool = Venv.get_pip_tool()
        if pip_tool == "uv":
            return ["uv", "pip", command, "--python", context.env_exe]
        return [pip_tool, command]

    @staticmethod
    def install_deps(context: types.SimpleNamespace, dependencies: List[str]) -> None:
        """Install a set of `dependencies` into Venv context using current pip tool. Makes
        sure that the right command context and arguments are used.
        """
        pip_tool = Venv.get_pip_tool()
        if pip_tool == "uv":
            cmd = [
                "uv",
                "pip",
                "install",
                "--python",
                context.env_exe,
                "--prerelease",
                "if-necessary-or-explicit",
            ]
            # on windows we need to copy the dependencies to the venv instead of linking
            if os.name == "nt":
                cmd.extend(["--link-mode", "copy"])
        else:
            cmd = [context.env_exe, "-Im", pip_tool, "install"]

        try:
            subprocess.check_output(
                cmd + dependencies, stderr=subprocess.STDOUT, errors="backslashreplace"
            )
        except subprocess.CalledProcessError as exc:
            raise CannotInstallDependencies(dependencies, context.env_exe, exc.output)

    @staticmethod
    def is_virtual_env() -> bool:
        """Checks if we are running in virtual environment"""
        return "VIRTUAL_ENV" in os.environ

    @staticmethod
    def is_venv_activated() -> bool:
        """Checks if virtual environment is activated in the shell"""
        return sys.prefix != sys.base_prefix
